from typing import List

import numpy as np
from scipy.spatial.transform import Rotation

import opengate_core as g4
from ..base import process_cls
from .base import ActorBase
from ..exception import fatal
from ..definitions import fwhm_to_sigma
from ..utility import g4_units
from ..image import (
    align_image_with_physical_volume,
    update_image_py_to_cpp,
    get_py_image_from_cpp_image,
)
from .actoroutput import ActorOutputRoot, ActorOutputSingleImage


def ene_win_peak(name, energy, energy_width_percent):
    energy_width_percent = energy * energy_width_percent / 2
    e_min = energy - energy_width_percent
    e_max = energy + energy_width_percent
    a = {"name": name, "min": e_min, "max": e_max}
    return a


def ene_win_down_scatter(name, peak_min_value, energy_width_percent):
    e_max = peak_min_value
    e_min = e_max / (1 + energy_width_percent / 2)
    a = {"name": name, "min": e_min, "max": e_max}
    return a


def ene_win_up_scatter(name, peak_max_value, energy_width_percent):
    e_min = peak_max_value
    e_max = e_min / (1 - energy_width_percent / 2)
    a = {"name": name, "min": e_min, "max": e_max}
    return a


def energy_windows_peak_scatter(
    peak_name,
    down_scatter_name,
    up_scatter_name,
    peak,
    peak_width,
    down_scatter_width,
    up_scatter_width=None,
):
    if up_scatter_width is None:
        up_scatter_width = down_scatter_width
    p = ene_win_peak(peak_name, peak, peak_width)
    do = ene_win_down_scatter(down_scatter_name, p["min"], down_scatter_width)
    up = ene_win_up_scatter(up_scatter_name, p["max"], up_scatter_width)
    return do, p, up


def get_simplified_digitizer_channels_Tc99m(spect_name, scatter_flag):
    keV = g4_units.keV
    # Tc99m
    channels = [
        {"name": f"scatter_{spect_name}", "min": 114 * keV, "max": 126 * keV},
        {"name": f"peak140_{spect_name}", "min": 126 * keV, "max": 154 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
    return channels


def get_simplified_digitizer_channels_Lu177(spect_name, scatter_flag):
    # Lu177, Ljungberg2016
    keV = g4_units.keV
    channels = [
        {"name": f"scatter1_{spect_name}", "min": 96 * keV, "max": 104 * keV},
        {"name": f"peak113_{spect_name}", "min": 104 * keV, "max": 121.48 * keV},
        {"name": f"scatter2_{spect_name}", "min": 122.48 * keV, "max": 133.12 * keV},
        {"name": f"scatter3_{spect_name}", "min": 176.46 * keV, "max": 191.36 * keV},
        {"name": f"peak208_{spect_name}", "min": 192.36 * keV, "max": 223.6 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
        channels.pop(1)
        channels.pop(1)
    return channels


def get_simplified_digitizer_channels_In111(spect_name, scatter_flag):
    # In111
    keV = g4_units.keV
    channels = [
        {"name": f"scatter1_{spect_name}", "min": 150 * keV, "max": 156 * keV},
        {"name": f"peak171_{spect_name}", "min": 156 * keV, "max": 186 * keV},
        {"name": f"scatter2_{spect_name}", "min": 186 * keV, "max": 192 * keV},
        {"name": f"scatter3_{spect_name}", "min": 218 * keV, "max": 224 * keV},
        {"name": f"peak245_{spect_name}", "min": 224 * keV, "max": 272 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
        channels.pop(1)
        channels.pop(1)
    return channels


def get_simplified_digitizer_channels_I131(spect_name, scatter_flag):
    # I131
    keV = g4_units.keV
    channels = [
        {"name": f"scatter1_{spect_name}", "min": 314 * keV, "max": 336 * keV},
        {"name": f"peak364_{spect_name}", "min": 336 * keV, "max": 392 * keV},
        {"name": f"scatter2_{spect_name}", "min": 392 * keV, "max": 414 * keV},
        {"name": f"scatter3_{spect_name}", "min": 414 * keV, "max": 556 * keV},
        {"name": f"scatter4_{spect_name}", "min": 556 * keV, "max": 595 * keV},
        {"name": f"peak637_{spect_name}", "min": 595 * keV, "max": 679 * keV},
    ]
    if not scatter_flag:
        channels.pop(0)
        channels.pop(1)
        channels.pop(1)
        channels.pop(1)
    return channels


def get_simplified_digitizer_channels_rad(spect_name, rad, scatter_flag):
    available_rad = {
        "Tc99m": get_simplified_digitizer_channels_Tc99m,
        "Lu177": get_simplified_digitizer_channels_Lu177,
        "In111": get_simplified_digitizer_channels_In111,
        "I131": get_simplified_digitizer_channels_I131,
    }

    if rad not in available_rad:
        fatal(
            f"Error, the radionuclide {rad} is not known, list of available is: {available_rad}"
        )

    return available_rad[rad](spect_name, scatter_flag)


class Digitizer:
    """
    Simple helper class to reduce the code size when creating a digitizer.
    It only avoids repeating attached_to, output and input_digi_collection parameters.
    """

    def __init__(self, sim, volume_name, digit_name):
        # input param
        self.simulation = sim
        self.volume_name = volume_name
        self.name = digit_name
        # store
        self.actors = []

        # start by the hit collection
        self.hc = self.set_hit_collection()

    def __str__(self):
        s = ""
        for a in self.actors:
            s += a.name + " "
        return s

    def set_hit_collection(self):
        hc = self.simulation.add_actor(
            "DigitizerHitsCollectionActor", f"{self.name}_hits"
        )
        hc.attached_to = self.volume_name
        hc.output_filename = ""
        hc.attributes = [
            "PostPosition",
            "TotalEnergyDeposit",
            "PreStepUniqueVolumeID",
            "PostStepUniqueVolumeID",
            "GlobalTime",
        ]
        self.actors.append(hc)
        return hc

    def add_module(self, module_type, module_name=None):
        index = len(self.actors)
        if module_name is None:
            module_name = f"{self.name}_{index}"
        mod = self.simulation.add_actor(module_type, module_name)
        mod.attached_to = self.actors[index - 1].attached_to
        if "input_digi_collection" in mod.user_info:
            mod.input_digi_collection = self.actors[index - 1].name
        first_key = next(iter(mod.user_output))
        if module_type == "DigitizerProjectionActor":
            mod.user_output[first_key].set_write_to_disk(False)
        else:
            mod.user_output[first_key].write_to_disk = False
        self.actors.append(mod)
        return mod

    def get_last_module(self):
        return self.actors[-1]

    def find_module(self, s):
        """
        Find the first module that contains the s string
        """
        for m in self.actors:
            if s == m.name:
                return m
        # if not found, find the closest
        for m in self.actors:
            if s in m.name:
                return m
        return None

    def find_module_by_type(this, module_type):
        for m in this.actors:
            if m.type_name == module_type:
                return m
        fatal(
            f'Error, the module type "{module_type}" is not found in the digitizer "{this.name}"'
        )
        return None


class DigitizerBase(ActorBase):
    _output_name_root = "root_output"

    # hints for IDE
    authorize_repeated_volumes: bool

    user_info_defaults = {
        "authorize_repeated_volumes": (
            False,
            {
                "doc": "User must say explicitly that the digitizer can work with repeated volumes",
            },
        ),
    }

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(self, *args, **kwargs)

    # def _add_user_output_root(self, **kwargs):
    #     """Specialized method to add a root user output in digitizers.
    #     The output name is hard-coded at the class-level and the same for all digitizers,
    #     i.e. in all digitizers, the user can do:
    #     digitizer.user_output.root_output
    #     Additionally, the C++ classes expect an output with this name.
    #     """
    #
    #     if self._output_name_root in self.user_output:
    #         fatal(
    #             f"The actor '{self.name}' already has a user_output called '{self._output_name_root}'."
    #             f"Probably, the method _add_user_output_root() was called more than once, "
    #             f"while it can be used only to add a single root output as in most digitizers. "
    #         )
    #     return self._add_user_output(ActorOutputRoot, self._output_name_root, **kwargs)

    def initialize(self):
        ActorBase.initialize(self)
        if self.authorize_repeated_volumes is True:
            return
        att = self.attached_to
        if not isinstance(self.attached_to, (list, tuple)):
            att = [self.attached_to]
        for a in att:
            current = self.simulation.volume_manager.get_volume(a).parent
            while current.name != "world" and hasattr(current, "g4_transform"):
                if len(current.g4_transform) > 1:
                    fatal(
                        f"This digitizer actor name '{self.name}' is attached to the volume '{self.attached_to}'. "
                        f"However, this volume is a daughter of the repeated volume '{current.name}'. It means it will "
                        f"gather data from all repeated instances. If you are "
                        f"sure, enable the option 'authorize_repeated_volumes'."
                    )
                current = current.parent


class DigitizerWithRootOutput(DigitizerBase):

    user_output_config = {
        "root_output": {
            "actor_output_class": ActorOutputRoot,
        },
    }


class DigitizerAdderActor(DigitizerWithRootOutput, g4.GateDigitizerAdderActor):
    """Equivalent to Gate "adder": gather all hits of an event in the same volume.
    Input: a HitsCollection, need aat least TotalEnergyDeposit and PostPosition attributes
    Output: a Single collections

    Policies:
    - EnergyWinnerPosition: consider position and energy of the hit with the max energy for all other attributes (Time, etc.): the value of the winner is used.
    - EnergyWeightedCentroidPosition: computed the energy-weighted centroid position for all other attributes (Time, etc.): the value the last seen hit is used.
    """

    user_info_defaults = {
        "attributes": (
            [],
            {
                "doc": "Attributes to be considered. ",
            },
        ),
        "output": (
            "singles.root",
            {
                "deprecated": "Use output_filename instead. ",
            },
        ),
        "input_digi_collection": (
            "Hits",
            {
                "doc": "Digi collection to be used as input. ",
            },
        ),
        "policy": (
            "EnergyWinnerPosition",
            {
                "doc": "Digi collection to be used as input. ",
                "allowed_values": (
                    "EnergyWeightedCentroidPosition",
                    "EnergyWinnerPosition",
                ),
            },
        ),
        "time_difference": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "number_of_hits": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "skip_attributes": (
            [],
            {
                "doc": "FIXME",
            },
        ),
        "clear_every": (
            1e5,
            {
                "doc": "FIXME",
            },
        ),
        "group_volume": (
            None,
            {
                "doc": "FIXME",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        DigitizerBase.__init__(self, *args, **kwargs)
        # self._add_user_output_root()
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateDigitizerAdderActor.__init__(self, self.user_info)
        self.AddActions({"StartSimulationAction", "EndSimulationAction"})

    def initialize(self):
        if (
            self.policy != "EnergyWinnerPosition"
            and self.policy != "EnergyWeightedCentroidPosition"
        ):
            fatal(
                f"Error, the policy for the Adder '{self.name}' must be EnergyWinnerPosition or "
                f"EnergyWeightedCentroidPosition, while is is '{self.policy}'"
            )
        DigitizerBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def set_group_by_depth(self):
        depth = -1
        if self.user_info.group_volume is not None:
            depth = self.simulation.volume_manager.get_volume(
                self.user_info.group_volume
            ).volume_depth_in_tree
        self.SetGroupVolumeDepth(depth)

    def StartSimulationAction(self):
        self.set_group_by_depth()
        DigitizerBase.StartSimulationAction(self)
        g4.GateDigitizerAdderActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerAdderActor.EndSimulationAction(self)


class DigitizerBlurringActor(DigitizerWithRootOutput, g4.GateDigitizerBlurringActor):
    """
    Digitizer module for blurring an attribute (single value only, not a vector).
    Usually for energy or time.
    """

    user_info_defaults = {
        "input_digi_collection": (
            "Hits",
            {
                "doc": "Digi collection to be used as input. ",
            },
        ),
        "blur_attribute": (
            None,
            {
                "doc": "FIXME",
            },
        ),
        "skip_attributes": (
            [],
            {
                "doc": "FIXME",
            },
        ),
        "clear_every": (
            1e5,
            {
                "doc": "FIXME",
            },
        ),
        "blur_method": (
            "Gaussian",
            {
                "doc": "FIXME",
                "allowed_values": (
                    "Gaussian",
                    "InverseSquare",
                    "Linear",
                ),
            },
        ),
        "blur_sigma": (
            None,
            {
                "doc": "FIXME",
            },
        ),
        "blur_fwhm": (
            None,
            {
                "doc": "FIXME",
            },
        ),
        "blur_reference_value": (
            0,
            {
                "doc": "FIXME",
            },
        ),
        "blur_resolution": (
            0,
            {
                "doc": "FIXME",
            },
        ),
        "blur_slope": (
            0,
            {
                "doc": "FIXME",
            },
        ),
    }

    type_name = "DigitizerBlurringActor"

    def __init__(self, *args, **kwargs):
        DigitizerBase.__init__(self, *args, **kwargs)
        # self._add_user_output_root()
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateDigitizerBlurringActor.__init__(self, self.user_info)
        self.AddActions({"StartSimulationAction", "EndSimulationAction"})

    def initialize(self):
        self.initialize_blurring_parameters()
        DigitizerBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def initialize_blurring_parameters(self):
        if self.blur_method == "Gaussian":
            self.set_param_gauss()
        if self.blur_method == "InverseSquare":
            self.set_param_inverse_square()
        if self.blur_method == "Linear":
            self.set_param_linear()

    def set_param_gauss(self):
        if self.blur_fwhm is not None and self.blur_sigma is not None:
            fatal(
                f"Error, use blur_sigma or blur_fwhm, not both "
                f"(there are: {self.blur_sigma} and {self.blur_fwhm}"
            )
        if self.blur_fwhm is not None:
            self.blur_sigma = self.blur_fwhm * fwhm_to_sigma
        if self.blur_sigma is None:
            fatal(f"Error, use blur_sigma or blur_fwhm")
        self.blur_reference_value = -1
        self.blur_resolution = -1
        self.blur_slope = 0

    def set_param_inverse_square(self):
        if self.blur_reference_value < 0 or self.blur_reference_value is None:
            fatal(
                f"Error, use positive blur_reference_value "
                f"(current value =  {self.blur_reference_value}"
            )
        if self.blur_resolution < 0 or self.blur_resolution is None:
            fatal(
                f"Error, use positive blur_resolution "
                f"(current value =  {self.blur_resolution}"
            )
        self.blur_fwhm = -1
        self.blur_sigma = -1
        if self.blur_slope is None:
            self.blur_slope = 0

    def set_param_linear(self):
        self.set_param_inverse_square()
        if self.blur_slope is None:
            fatal(
                f"Error, use positive blur_slope "
                f"(current value =  {self.blur_slope}"
            )

    def StartSimulationAction(self):
        DigitizerBase.StartSimulationAction(self)
        g4.GateDigitizerBlurringActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerBlurringActor.EndSimulationAction(self)


class DigitizerSpatialBlurringActor(
    DigitizerWithRootOutput, g4.GateDigitizerSpatialBlurringActor
):
    """
    Digitizer module for blurring a (global) spatial position.
    """

    # hints for IDE
    attributes: List
    input_digi_collection: str
    skip_attributes: List
    clear_every: int
    blur_attribute: str
    blur_fwhm: float
    blur_sigma: float
    keep_in_solid_limits: bool
    use_truncated_Gaussian: bool

    user_info_defaults = {
        "attributes": (
            [],
            {
                "doc": "Attributes to be considered. ",
            },
        ),
        "input_digi_collection": (
            "Hits",
            {
                "doc": "FIXME",
            },
        ),
        "skip_attributes": (
            [],
            {
                "doc": "FIXME",
            },
        ),
        "clear_every": (
            int(1e5),
            {
                "doc": "FIXME",
            },
        ),
        "blur_attribute": (
            None,
            {
                "doc": "FIXME",
            },
        ),
        "blur_fwhm": (
            None,
            {
                "doc": "FIXME",
            },
        ),
        "blur_sigma": (
            None,
            {
                "doc": "FIXME",
            },
        ),
        "keep_in_solid_limits": (
            True,
            {
                "doc": "FIXME",
            },
        ),
        "use_truncated_Gaussian": (
            True,
            {
                "doc": "FIXME",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        # base classes
        ActorBase.__init__(self, *args, **kwargs)
        # self._add_user_output_root()
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateDigitizerSpatialBlurringActor.__init__(self, self.user_info)
        self.AddActions({"StartSimulationAction", "EndSimulationAction"})

    def initialize_blurring_parameters(self):
        if self.blur_fwhm is not None and self.blur_sigma is not None:
            fatal(
                f"Error, use blur_sigma or blur_fwhm, not both "
                f"(there are: {self.blur_sigma} and {self.blur_fwhm}"
            )
        if not hasattr(self.blur_sigma, "__len__"):
            self.blur_sigma = [self.blur_sigma] * 3
        if not hasattr(self.blur_fwhm, "__len__"):
            self.blur_fwhm = [self.blur_fwhm] * 3
        if self.blur_fwhm is not None:
            self.blur_sigma = np.array(self.blur_fwhm) * fwhm_to_sigma
        if self.blur_sigma is None:
            fatal(f"Error, use blur_sigma or blur_fwhm")

    def initialize(self):
        self.initialize_blurring_parameters()
        DigitizerBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def StartSimulationAction(self):
        DigitizerBase.StartSimulationAction(self)
        g4.GateDigitizerSpatialBlurringActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerSpatialBlurringActor.EndSimulationAction(self)


class DigitizerEfficiencyActor(
    DigitizerWithRootOutput, g4.GateDigitizerEfficiencyActor
):
    """
    Digitizer module for simulating efficiency.
    """

    user_info_defaults = {
        "attributes": (
            [],
            {
                "doc": "Attributes to be considered. ",
            },
        ),
        "input_digi_collection": (
            "Hits",
            {
                "doc": "FIXME",
            },
        ),
        "skip_attributes": (
            [],
            {
                "doc": "FIXME",
            },
        ),
        "clear_every": (
            1e5,
            {
                "doc": "FIXME",
            },
        ),
        "efficiency": (
            1.0,
            {
                "doc": "FIXME",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        # base classes
        ActorBase.__init__(self, *args, **kwargs)
        # self._add_user_output_root()
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateDigitizerEfficiencyActor.__init__(self, self.user_info)
        self.AddActions({"StartSimulationAction", "EndSimulationAction"})

    def initialize_blurring_parameters(self):
        if not (0.0 <= self.efficiency <= 1.0):
            self.warn_user(
                f"Efficency set to {self.efficiency}, which is not in [0;1]."
            )

    def initialize(self):
        self.initialize_blurring_parameters()
        DigitizerBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def StartSimulationAction(self):
        DigitizerBase.StartSimulationAction(self)
        g4.GateDigitizerEfficiencyActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerEfficiencyActor.EndSimulationAction(self)


class DigitizerEnergyWindowsActor(
    DigitizerWithRootOutput, g4.GateDigitizerEnergyWindowsActor
):
    """
    Consider a list of hits and arrange them according to energy intervals.
    Input: one DigiCollection
    Output: as many DigiCollections as the number of energy windows
    """

    user_info_defaults = {
        "attributes": (
            [],
            {
                "doc": "Attributes to be considered. ",
            },
        ),
        "input_digi_collection": (
            "Hits",
            {
                "doc": "FIXME",
            },
        ),
        "skip_attributes": (
            [],
            {
                "doc": "FIXME",
            },
        ),
        "clear_every": (
            1e5,
            {
                "doc": "FIXME",
            },
        ),
        "channels": (
            [],
            {
                "doc": "FIXME",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        DigitizerBase.__init__(self, *args, **kwargs)
        # self._add_user_output_root()
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateDigitizerEnergyWindowsActor.__init__(self, self.user_info)
        self.AddActions({"StartSimulationAction", "EndSimulationAction"})

    def initialize(self):
        DigitizerBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def StartSimulationAction(self):
        DigitizerBase.StartSimulationAction(self)
        g4.GateDigitizerEnergyWindowsActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        DigitizerBase.EndSimulationAction(self)
        g4.GateDigitizerEnergyWindowsActor.EndSimulationAction(self)


class DigitizerHitsCollectionActor(
    DigitizerWithRootOutput, g4.GateDigitizerHitsCollectionActor
):
    """
    Build a list of hits in a given volume.
    - the list of attributes to be stored is given in the 'attributes' options
    - output as root
    """

    user_info_defaults = {
        "attributes": (
            [],
            {
                "doc": "Attributes to be considered. ",
            },
        ),
        "clear_every": (
            1e5,
            {
                "doc": "FIXME",
            },
        ),
        "debug": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "keep_zero_edep": (
            False,
            {
                "doc": "FIXME",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        DigitizerBase.__init__(self, *args, **kwargs)
        # self._add_user_output_root()
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateDigitizerHitsCollectionActor.__init__(self, self.user_info)
        self.AddActions({"StartSimulationAction", "EndSimulationAction"})

    def initialize(self):
        DigitizerBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def StartSimulationAction(self):
        DigitizerBase.StartSimulationAction(self)
        g4.GateDigitizerHitsCollectionActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        DigitizerBase.EndSimulationAction(self)
        g4.GateDigitizerHitsCollectionActor.EndSimulationAction(self)


class DigitizerProjectionActor(DigitizerBase, g4.GateDigitizerProjectionActor):
    """
    This actor takes as input HitsCollections and performed binning in 2D images.
    If there are several HitsCollection as input, the slices will correspond to each HC.
    If there are several runs, images will also be slice-stacked.
    """

    # hints for IDE
    input_digi_collections: List[str]
    spacing: List[float]
    size: List[int]
    physical_volume_index: int
    origin_as_image_center: bool
    detector_orientation_matrix: np.ndarray

    user_info_defaults = {
        # FIXME: implement a setter hook so the user can provided digitizer instances instead of their name,
        # like in attached_to
        "input_digi_collections": (
            ["Hits"],
            {
                "doc": "FIXME",
            },
        ),
        "spacing": (
            [4 * g4_units.mm, 4 * g4_units.mm],
            {"doc": "FIXME"},
        ),
        "size": (
            [128, 128],
            {"doc": "size of the 2D projection in pixels"},
        ),
        "physical_volume_index": (
            -1,
            {
                "doc": "When attached to a repeated volume, this option indicate which copy is used",
            },
        ),
        "origin_as_image_center": (
            True,
            {
                "doc": "FIXME",
            },
        ),
        "detector_orientation_matrix": (
            Rotation.from_euler("x", 0).as_matrix(),
            {
                "doc": "FIXME",
            },
        ),
    }

    user_output_config = {
        "counts": {
            "actor_output_class": ActorOutputSingleImage,
        },
        "squared_counts": {
            "actor_output_class": ActorOutputSingleImage,
            "active": False,
        },
    }

    def __init__(self, *args, **kwargs):
        DigitizerBase.__init__(self, *args, **kwargs)
        # self._add_user_output(ActorOutputSingleImage, "projection")
        self.start_output_origin = None
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateDigitizerProjectionActor.__init__(self, self.user_info)
        self.AddActions({"StartSimulationAction", "EndSimulationAction"})

    def initialize(self):
        # for the moment, we cannot use this actor with several volumes
        m = self.attached_to
        if hasattr(m, "__len__") and not isinstance(m, str):
            fatal(
                f"Sorry, cannot (yet) use several attached_to volumes for "
                f"DigitizerProjectionActor {self.user_info.name}"
            )
        if self.authorize_repeated_volumes is True:
            fatal(
                f"Sorry, cannot (yet) use ProjectionActor with repeated volumes, "
                f"set 'authorize_repeated_volumes' to False"
            )
        DigitizerBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    @property
    def output_size(self):
        # consider 3D images, third dimension can be the energy windows
        output_size = list(self.size)
        if len(output_size) != 2:
            fatal(
                f"Error, the size must be a 2-vector (2D) while it is {output_size}. "
                f"Note: The size along the third dimension is automatically set to 1."
            )
        output_size.append(1)
        return output_size

    @property
    def output_spacing(self):
        # consider 3D images, the third dimension can be the energy windows
        output_spacing = list(self.spacing)
        if len(output_spacing) != 2:
            fatal(
                f"Error, the spacing must be a 2-vector (2D) while it is {output_spacing}. "
                f"Note: The spacing along the third dimension is automatically set to 1."
            )
        output_spacing.append(1)
        return output_spacing

    def compute_thickness(self, volume, channels):
        """
        Get the thickness of the detector volume, in the correct direction.
        By default, it is Z. We use the 'projection_orientation' to get the correct one.
        """
        vol = self.actor_engine.simulation_engine.volume_engine.get_volume(volume)
        if len(vol.g4_physical_volumes) < 1:
            fatal(
                f"The actor {self.name} is attached to '{self.attached_to}' which has "
                f"no associated g4_physical_volumes (probably a parameterized volume?) and cannot "
                f"be used. Try to attach it to the mother volume of the parameterized volume."
            )

        solid = vol.g4_physical_volumes[0].GetLogicalVolume().GetSolid()
        pMin = g4.G4ThreeVector()
        pMax = g4.G4ThreeVector()
        solid.BoundingLimits(pMin, pMax)
        d = np.array([0, 0, 1.0])
        d = np.dot(self.detector_orientation_matrix, d)
        imax = np.argmax(d)
        thickness = (pMax[imax] - pMin[imax]) / channels
        return thickness

    def StartSimulationAction(self):
        DigitizerBase.StartSimulationAction(self)
        # for the moment, we cannot use this actor with several volumes
        if hasattr(self.attached_to, "__len__") and not isinstance(
            self.attached_to, str
        ):
            fatal(
                f"Sorry, cannot (yet) use several mothers volumes for "
                f"DigitizerProjectionActor {self.name}"
            )

        # define the new size and spacing according to the nb of channels
        # and according to the volume shape
        size = self.output_size
        spacing = self.output_spacing
        size[2] = len(self.input_digi_collections) * len(
            self.simulation.run_timing_intervals
        )
        spacing[2] = self.compute_thickness(self.attached_to, size[2])

        # we use the image associated with run 0 for the entire simulation
        # in the future, this actor should implement a BeginOfRunActionMasterThread
        # to be able to work on a per-run basis
        self.user_output.counts.create_empty_image(0, size, spacing)

        # check physical_volume_index and number of repeating
        n = len(self.attached_to_volume.g4_physical_volumes)
        if n > 1:
            if self.physical_volume_index >= n or self.physical_volume_index < 0:
                fatal(
                    f"The actor '{self.name}' is attached to '{self.attached_to}' which is repeated {n} times. "
                    f"You must set a valid 'physical_volume_index' between O to {n-1} while "
                    f"it is {self.physical_volume_index}."
                )
        else:
            # force the index to be zero
            self.physical_volume_index = 0

        # initial position (will be anyway updated in BeginOfRunSimulation)
        try:
            pv = self.attached_to_volume.g4_physical_volumes[self.physical_volume_index]
        except KeyError:
            fatal(
                f"Error in the DigitizerProjectionActor {self.name}. "
                f"No physical volume found for index {self.physical_volume_index} "
                f"in volume {self.attached_to_volume.name}"
            )
            pv = None  # avoid warning from IDE
        align_image_with_physical_volume(
            self.attached_to_volume, self.user_output.counts.data_per_run[0].image
        )
        self.SetPhysicalVolumeName(str(pv.GetName()))

        # update the cpp image and start
        update_image_py_to_cpp(
            self.user_output.counts.data_per_run[0].image, self.fImage, True
        )

        # uncertainty ?
        if self.user_output.squared_counts.get_active():
            self.user_output.squared_counts.create_empty_image(0, size, spacing)
            align_image_with_physical_volume(
                self.attached_to_volume,
                self.user_output.squared_counts.data_per_run[0].image,
            )
            update_image_py_to_cpp(
                self.user_output.squared_counts.data_per_run[0].image,
                self.fSquaredImage,
                True,
            )
            self.EnableSquaredImage(True)

        # keep the initial origin
        self.start_output_origin = list(
            self.user_output.counts.data_per_run[0].get_image_properties()[0].origin
        )
        g4.GateDigitizerProjectionActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerProjectionActor.EndSimulationAction(self)

        # retrieve the image
        self.user_output.counts.store_data(
            "merged", get_py_image_from_cpp_image(self.fImage)
        )

        # set its properties
        info = self.user_output.counts.data_per_run[0].get_image_properties()[0]
        spacing = info.spacing
        if self.origin_as_image_center:
            origin = -info.size * spacing / 2.0 + spacing / 2.0
        else:
            origin = self.start_output_origin
        origin[2] = 0
        spacing[2] = 1
        self.user_output.counts.merged_data.SetSpacing(list(spacing))
        self.user_output.counts.merged_data.SetOrigin(list(origin))

        # remove the image for run 0 as the result is in merged_data
        self.user_output.counts.data_per_run.pop(0)

        self.user_output.counts.write_data_if_requested(which="merged")

        # squared ?
        if self.user_output.squared_counts.get_active():
            self.user_output.squared_counts.store_data(
                "merged", get_py_image_from_cpp_image(self.fSquaredImage)
            )
            self.user_output.squared_counts.merged_data.SetSpacing(list(spacing))
            self.user_output.squared_counts.merged_data.SetOrigin(list(origin))
            self.user_output.squared_counts.data_per_run.pop(0)
            self.user_output.squared_counts.write_data_if_requested(which="merged")


class DigitizerReadoutActor(DigitizerAdderActor, g4.GateDigitizerReadoutActor):
    """
    This actor is a DigitizerAdderActor + a discretization step:
    the final position is the center of the volume
    """

    user_info_defaults = {
        "discretize_volume": (
            None,
            {
                "doc": "FIXME",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        # warning : inherit from DigitizerAdderActor but should not use its
        # constructor because it adds an output
        ActorBase.__init__(self, *args, **kwargs)
        # self._add_user_output_root()
        self.__initcpp__()

    def __initcpp__(self):
        # python 3.8 complains about init not called, we add explicit call to
        # GateDigitizerAdderActor here (not needed for py > 3.8)
        g4.GateDigitizerAdderActor.__init__(self, self.user_info)
        g4.GateDigitizerReadoutActor.__init__(self, self.user_info)
        self.AddActions({"StartSimulationAction", "EndSimulationAction"})

    def StartSimulationAction(self):
        DigitizerBase.StartSimulationAction(self)
        DigitizerAdderActor.set_group_by_depth(self)
        if self.user_info.discretize_volume is None:
            fatal('Please, set the option "discretize_volume"')
        depth = self.simulation.volume_manager.get_volume(
            self.discretize_volume
        ).volume_depth_in_tree
        self.SetDiscretizeVolumeDepth(depth)
        g4.GateDigitizerReadoutActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerReadoutActor.EndSimulationAction(self)


class PhaseSpaceActor(DigitizerWithRootOutput, g4.GatePhaseSpaceActor):
    """
    Similar to HitsCollectionActor : store a list of hits.
    The hits can be stored when entering, exiting, first time seen or always (all hits stored).
    """

    user_info_defaults = {
        "attributes": (
            [],
            {
                "doc": "FIXME",
            },
        ),
        "store_absorbed_event": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "debug": (
            False,
            {
                "doc": "print debug info",
            },
        ),
        "steps_to_store": (
            "entering",
            {
                "doc": "Define when to store the hits, can be 'entering', 'exiting', 'first' or 'all'. "
                "Or several values separated by a space. ",
                # "allowed_values": ["entering", "exiting", "first", "all"], # can be multiple
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        DigitizerBase.__init__(self, *args, **kwargs)
        # self._add_user_output_root()
        self.total_number_of_entries = 0
        self.number_of_absorbed_events = 0
        self.__initcpp__()

    def __initcpp__(self):
        g4.GatePhaseSpaceActor.__init__(self, self.user_info)

    def initialize(self):
        DigitizerBase.initialize(self)
        if "entering" in self.steps_to_store:
            self.SetStoreEnteringStepFlag(True)
        if "exiting" in self.steps_to_store:
            self.SetStoreExitingStepFlag(True)
        if "first" in self.steps_to_store:
            self.SetStoreFirstStepInVolumeFlag(True)
        if "all" in self.steps_to_store:
            self.SetStoreAllStepsFlag(True)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def StartSimulationAction(self):
        DigitizerBase.StartSimulationAction(self)
        g4.GatePhaseSpaceActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        self.number_of_absorbed_events = self.GetNumberOfAbsorbedEvents()
        self.total_number_of_entries = self.GetTotalNumberOfEntries()
        if self.total_number_of_entries == 0:
            self.warn_user(
                f"Empty output, no particles stored in {self.get_output_path()}"
            )
        g4.GatePhaseSpaceActor.EndSimulationAction(self)


class DigiAttributeProcessDefinedStepInVolumeActor(
    ActorBase, g4.GateDigiAttributeProcessDefinedStepInVolumeActor
):
    """
    This actor is use when the user create a ProcessDefinedStepAttribute.
    The actor is automatically created and use to store how many time a given process (process_name)
    occur in a given volume (attached_to).
    This actor is not intended to be used directly by the user.
    """

    user_info_defaults = {
        "process_name": (None, {}),
    }

    def __init__(self, *args, **kwargs):
        ActorBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateDigiAttributeProcessDefinedStepInVolumeActor.__init__(
            self, self.user_info
        )

    def initialize(self):
        ActorBase.initialize(self)
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def StartSimulationAction(self):
        ActorBase.StartSimulationAction(self)
        g4.GateDigiAttributeProcessDefinedStepInVolumeActor.StartSimulationAction(self)


class ProcessDefinedStepInVolumeAttribute:
    """ """

    def __init__(self, sim, process_name, volume_name):
        self.name = f"ProcessDefinedStep__{process_name}__{volume_name}"
        self.actor = sim.add_actor(
            "DigiAttributeProcessDefinedStepInVolumeActor", self.name
        )
        self._process_name = process_name
        self._volume_name = volume_name
        self.actor.attached_to = volume_name
        self.actor.process_name = process_name
        # self.actor.priority = 1  # FIXME before other

    @property
    def process_name(self):
        return self._process_name

    @property
    def volume_name(self):
        return self._volume_name

    @process_name.setter
    def process_name(self, value):
        fatal(f"Cannot change dynamically the process name")

    @volume_name.setter
    def volume_name(self, value):
        fatal(f"Cannot change dynamically the volume name")


process_cls(DigitizerBase)
process_cls(DigitizerWithRootOutput)
process_cls(DigitizerAdderActor)
process_cls(DigitizerBlurringActor)
process_cls(DigitizerSpatialBlurringActor)
process_cls(DigitizerEfficiencyActor)
process_cls(DigitizerEnergyWindowsActor)
process_cls(DigitizerHitsCollectionActor)
process_cls(DigitizerProjectionActor)
process_cls(DigitizerReadoutActor)
process_cls(PhaseSpaceActor)
process_cls(DigiAttributeProcessDefinedStepInVolumeActor)



import tables
from pathlib import Path

from ..exception import fatal
from .utility import get_nodes_in_group, get_node_name
from .unitbase import ProcessingUnitBase


def _setter_hook_input_path(self, p):
    if self.user_info["actor_output"] is None:
        return Path(p)
    else:
        fatal(
            f"Cannot set 'input_path' and 'actor_output'. Unset one before setting the other. "
        )


def _setter_hook_input_actor(self, input_actor):
    if self.user_info["input_actor"] is None:
        try:
            n = input_actor.name
        except AttributeError:
            n = input_actor
        return n
    else:
        fatal(
            f"Cannot set 'input_path' and 'input_actor'. Unset one before setting the other. "
        )


class DataFetcherBase(ProcessingUnitBase):
    """This is a base class for data fetchers."""

    user_info_defaults = {
        "input_path": (
            None,
            {
                "doc": "Path to the input file. ",
                "setter_hook": _setter_hook_input_path,
            },
        ),
        "actor_output": (
            None,
            {
                "doc": "Actor that provides data. ",
                "setter_hook": _setter_hook_input_actor,
            },
        ),
        "which_output": (
            "merged_data",
            {
                "doc": "Which kind of output should be taken from the actor? "
                "Allowed values: 'merged_data' or a valid run index.",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._input_path = None  # internal variable to store input path

    @property
    def can_be_initial_unit(self):
        return True

    def _get_input_path(self):
        if self.input_path is not None:
            return Path(self.input_path).absolute()
        elif self.actor_output is not None:
            return self.actor_output.get_output_path(self.which_output)
        else:
            return None

    def close(self):
        super().close()
        self._input_path = None


class DataFetcherWithFileHandle(DataFetcherBase):

    def _open_input_file_handle(self):
        input_path = self._get_input_path()
        if self._input_path is None:
            if not self.post_processor.has_extra_file_handle(input_path):
                self.post_processor.get_or_register_extra_file_handle(
                    self.open_file(input_path)
                )
            self._input_path = input_path

    def open_file(self, path):
        """Needs concrete implementation in derived class."""
        raise NotImplementedError

    @property
    def input_file_handle(self):
        return self.post_processor.get_extra_file_handle(self._input_path)


def _setter_hook_path_in_hdf5_file(self, p):
    if not p.startswith("/"):
        p = "/" + p
    return p


class DataFetcherHdf5(DataFetcherWithFileHandle):
    user_info_defaults = {
        # "hdf5_node_name": (
        #     None,
        #     {
        #         "doc": "Name of the attribute to be blurred.",
        #     },
        # ),
        "input_hdf5_group": (
            "/",
            {
                "doc": "Name of the attribute to be blurred.",
                "setter_hook": _setter_hook_path_in_hdf5_file,
            },
        ),
    }

    def open_file(self, path):
        return tables.open_file(path, mode="r", title=self.get_title_for_hdf5_node())

    def do_your_job(self):
        self._open_input_file_handle()
        input_group = self.input_file_handle.get_node(
            "/", name=self.input_hdf5_group.lstrip("/")
        )
        nodes = get_nodes_in_group(input_group, node_name=self.input_name)
        print("Nodes in DataFetcherHdf5.do_your_job()")
        for n in nodes:
            print(n)
            self.register_output_data_handle(
                self.post_processor.create_external_link(
                    self.get_unit_output_group(), get_node_name(n), target=n
                )
            )



import itk
import numpy as np
import json
from box import Box

from ..exception import fatal, warning, GateImplementationError
from ..utility import ensure_filename_is_str, calculate_variance
from ..image import (
    sum_itk_images,
    divide_itk_images,
    multiply_itk_images,
    scale_itk_image,
    create_3d_image,
    write_itk_image,
    get_info_from_image,
    itk_image_from_array,
    add_constant_to_itk_image,
)


class DataItem:
    """This is the base class for all data items.
    It stores the actual data, e.g. an array, an image, etc. in an attribute 'data'.

    Derived classes can (should) implement merge_with and inplace_merge_with
    so actor output using this data container with the respective data items can be merged after runs.

    Derived classes should also implement an appropriate write method
    if data writing is supposed to be handled on the python-side.

    Derived classes can also implement arithmetic operator like __add__, __mul__, etc.
    """

    def __init__(self, *args, data=None, meta_data=None, **kwargs):
        self.data = None
        if data is not None:
            self.set_data(data)
        self.meta_data = Box({"number_of_samples": 1})
        if meta_data:
            try:
                for k, v in meta_data.items():
                    self.meta_data[k] = v
            except AttributeError:
                fatal(
                    f"Illegal keyword argument meta_data: {meta_data}. "
                    f"Should be a dictionary-like object, but found {type(meta_data)}."
                )

    def set_data(self, data, **kwargs):
        self.data = data

    @property
    def data_is_none(self):
        return self.data is None

    def _assert_data_is_not_none(self):
        if self.data_is_none:
            raise ValueError(
                "This data item does not contain any data yet. "
                "Use set_data() before applying any operations. "
            )

    def __add__(self, other):
        return NotImplemented

    def __iadd__(self, other):
        return NotImplemented

    def __mul__(self, other):
        return NotImplemented

    def __imul__(self, other):
        return NotImplemented

    def __truediv__(self, other):
        return NotImplemented

    def __getattr__(self, item):
        # check if any of the data items has this attribute
        # exclude 'data' to avoid infinite recursion
        # exclude '__setstate__' and '__getstate__' to avoid interference with pickling
        if item not in ("data", "__setstate__", "__getstate__"):
            if hasattr(self.data, item):
                if callable(getattr(self.data, item)):

                    def hand_down(*args, **kwargs):
                        getattr(self.data, item)(*args, **kwargs)

                    return hand_down
            else:
                raise AttributeError(f"No such attribute '{item}'")
        else:
            raise AttributeError(f"No such attribute '{item}'")

    def merge_with(self, other):
        """The base class does not implement merging.
        Specific classes can override this, e.g. to merge mean values.
        """
        raise NotImplementedError(
            f"Method 'inplace_merge_with' not implemented for data item class {type(self)} "
        )

    def inplace_merge_with(self, other):
        """The base class does not implement merging.
        Specific classes can override this, e.g. to merge mean values.
        """
        raise NotImplementedError(
            f"Method 'inplace_merge_with' not implemented for data item class {type(self)} "
        )

    def write(self, *args, **kwargs):
        raise NotImplementedError(f"This is the base class. ")

    @property
    def number_of_samples(self):
        try:
            return self.meta_data["number_of_samples"]
        except KeyError:
            fatal(
                f"This data item holds a mean value, "
                f"but the meta_data dictionary does not contain any value for 'number_of_samples'."
            )

    @number_of_samples.setter
    def number_of_samples(self, value):
        self.meta_data["number_of_samples"] = int(value)


class MeanValueDataItemMixin:
    """This class cannot be instantiated on its own.
    It is solely meant to be mixed into a class that inherits from DataItem (or daughters).
    Important: It must appear before the main base class in the inheritance order so that the
    overloaded methods take priority.
    """

    # hints for IDE
    number_of_samples: int

    def merge_with(self, other):
        result = (self * self.number_of_samples + other * other.number_of_samples) / (
            self.number_of_samples + other.number_of_samples
        )
        result.number_of_samples = self.number_of_samples + other.number_of_samples
        return result

    def inplace_merge_with(self, other):
        if self.data is None:
            self.set_data(other.data)
            self.number_of_samples = other.number_of_samples
        else:
            self *= self.number_of_samples
            other *= other.number_of_samples
            self += other
            self /= self.number_of_samples + other.number_of_samples
            self.number_of_samples = self.number_of_samples + other.number_of_samples
        return self


class ArithmeticDataItem(DataItem):
    """Base class for data items where the data component already has implemented arithmetic operators.
    Examples: Scalars, Numpy arrays, etc.
    """

    def __iadd__(self, other):
        if self.data_is_none:
            raise ValueError(
                "This data item does not contain any data yet. "
                "Use set_data() before applying any operations. "
            )
        self.set_data(self.data + other.data)
        return self

    def __add__(self, other):
        if self.data_is_none:
            raise ValueError(
                "This data item does not contain any data yet. "
                "Use set_data() before applying any operations. "
            )
        return type(self)(data=self.data + other.data)

    def __mul__(self, other):
        if self.data_is_none:
            raise ValueError(
                "This data item does not contain any data yet. "
                "Use set_data() before applying any operations. "
            )
        return type(self)(data=self.data * other.data)

    def __imul__(self, other):
        if self.data_is_none:
            raise ValueError(
                "This data item does not contain any data yet. "
                "Use set_data() before applying any operations. "
            )
        self.set_data(self.data * other.data)
        return self

    def __truediv__(self, other):
        if self.data_is_none:
            raise ValueError(
                "This data item does not contain any data yet. "
                "Use set_data() before applying any operations. "
            )
        return type(self)(data=self.data / other.data)

    def __itruediv__(self, other):
        if self.data_is_none:
            raise ValueError(
                "This data item does not contain any data yet. "
                "Use set_data() before applying any operations. "
            )
        self.set_data(self.data / other.data)
        return self

    def write(self, path, **kwargs):
        np.savetxt(path, self.data)


# data items holding arrays
class ArrayDataItem(ArithmeticDataItem):

    def set_data(self, data):
        super().set_data(np.asarray(data))


class ScalarDataItem(ArithmeticDataItem):

    def write(self, *args, **kwargs):
        raise NotImplementedError


# data items holding images
class ItkImageDataItem(DataItem):

    @property
    def image(self):
        return self.data

    @property
    def image_array(self):
        return itk.array_view_from_image(self.image)

    def inplace_merge_with(self, other):
        if self.data is None:
            self.set_data(other.data)
            self.number_of_samples = other.number_of_samples
        else:
            self.__iadd__(other)
            self.number_of_samples += other.number_of_samples
        return self

    def __iadd__(self, other):
        self._assert_data_is_not_none()
        self.set_data(sum_itk_images([self.data, other.data]))
        return self

    def __add__(self, other):
        self._assert_data_is_not_none()
        if isinstance(other, (float, int)):
            return type(self)(data=add_constant_to_itk_image(self.data, other))
        else:
            return type(self)(data=sum_itk_images([self.data, other.data]))

    def __mul__(self, other):
        self._assert_data_is_not_none()
        if isinstance(other, (float, int)):
            return type(self)(data=scale_itk_image(self.data, other))
        else:
            return type(self)(data=multiply_itk_images([self.data, other.data]))

    def __imul__(self, other):
        self._assert_data_is_not_none()
        if isinstance(other, (float, int)):
            self.set_data(scale_itk_image(self.data, other))
        else:
            self.set_data(multiply_itk_images([self.data, other.data]))
        return self

    def __truediv__(self, other):
        self._assert_data_is_not_none()
        if isinstance(other, (float, int)):
            return type(self)(data=scale_itk_image(self.data, 1.0 / other))
        else:
            return type(self)(data=divide_itk_images(self.data, other.data))

    def __itruediv__(self, other):
        self._assert_data_is_not_none()
        if isinstance(other, (float, int)):
            self.set_data(scale_itk_image(self.data, 1.0 / other))
        else:
            self.set_data(divide_itk_images(self.data, other.data))
        return self

    def set_image_properties(self, **properties):
        if not self.data_is_none:
            if "spacing" in properties and properties["spacing"] is not None:
                self.data.SetSpacing(properties["spacing"])
            if "origin" in properties and properties["origin"] is not None:
                self.data.SetOrigin(properties["origin"])
            if "rotation" in properties and properties["rotation"] is not None:
                self.data.SetDirection(properties["rotation"])

    def get_image_properties(self):
        return get_info_from_image(self.data)

    def copy_image_properties(self, other_image):
        self.data.CopyInformation(other_image)

    def set_array_to_image(self, arr):
        image = itk_image_from_array(arr)
        image.SetOrigin(self.image.GetOrigin())
        image.SetSpacing(self.image.GetSpacing())
        image.SetDirection(self.image.GetDirection())
        self.data = image

    def create_empty_image(
        self,
        size,
        spacing,
        origin=None,
        pixel_type="float",
        allocate=True,
        fill_value=0,
    ):
        self.set_data(
            create_3d_image(size, spacing, origin, pixel_type, allocate, fill_value)
        )

    def write(self, path):
        write_itk_image(self.data, ensure_filename_is_str(path))


class MeanItkImageDataItem(MeanValueDataItemMixin, ItkImageDataItem):
    """This class represents an ITK image which is meant to hold mean values per voxel.
    The class MeanValueDataItemMixin therefore overloads the merge_with and inplace_merge_with methods.
    """


class DataContainer:
    """Common base class for all containers. Nothing implemented here for now."""

    def __init__(self, belongs_to, *args, **kwargs):
        self.belongs_to = belongs_to

    def __copy__(self):
        return type(self)(self.belongs_to)


class DataDictionary(DataContainer):

    def __init__(self, initial_dict, *args, encoder="json", **kwargs):
        self.data = dict([(k, v) for k, v in initial_dict.items()])
        available_encoders = ("json",)
        if encoder in available_encoders:
            self.encoder = encoder
        else:
            fatal(f"Invalid encoder. Available encoders are: {available_encoders}")

    def write(self, path):
        if self.encoder == "json":
            with open(path, "w") as f:
                json.dump(self.data, f, indent=4)


class DataItemContainer(DataContainer):
    """This is a base class. Inherit from it to implement specific containers."""

    # No data item classes specified in the base class.
    # Derived classes must specify this at the class level
    _data_item_classes = ()

    # let the class know which properties should be treated as data items
    __extra_data_items__ = ()

    def __init__(self, *args, data=None, **kwargs):
        super().__init__(*args, **kwargs)

        # create the instances of the data item classes
        # and populate them with data if provided
        self.data = [dic(data=None) for dic in self._data_item_classes]
        if data is not None:
            self.set_data(*data)

    def __copy__(self):
        obj = super().__copy__()
        obj.set_data(*self.data)
        return obj

    @classmethod
    def __get_data_item_names__(cls):
        data_item_names = [i for i in range(len(cls._data_item_classes))]
        data_item_names.extend(cls.__extra_data_items__)
        return data_item_names

    # the actual write config needs to be fetched from the actor output instance
    # which handles this data item container
    @property
    def data_item_config(self):
        try:
            return self.belongs_to.data_item_config
        except AttributeError:
            raise GateImplementationError("belongs_to unknown")

    @property
    def _tuple_length(self):
        return len(self._data_item_classes)

    @property
    def meta_data(self):
        if self._tuple_length > 1:
            return [d.meta_data for d in self.data]
        else:
            return self.data[0].meta_data

    @meta_data.setter
    def meta_data(self, meta_data):
        for d in self.data:
            d.meta_data = meta_data

    def update_meta_data(self, meta_data):
        for d in self.data:
            if d is not None:
                d.meta_data.update(meta_data)

    def set_data(self, *data, item=None):
        # data might be already contained in the correct container class,
        # or intended to be the input to the container class
        if item is not None:
            if len(data) != len(item):
                fatal(
                    f"Inconsistent input to set_data method: "
                    f"{len(data)} data items provided, "
                    f"but {len(item)} items specified in the 'item' keyword argument. "
                )
        else:
            item = [i for i in range(len(data))]
        processed_data = []
        for i, d in zip(item, data):
            c = self._data_item_classes[i]
            if isinstance(d, c):
                processed_data.append(d)
            else:
                processed_data.append(c(data=d))
        # Fill up the data list with None in case not all data were passed
        # processed_data.extend([None] * (len(self._data_item_classes) - len(data)))
        self.data = processed_data

    def get_data_item_object(self, item):
        try:
            identifier = int(item)
            try:
                return self.data[identifier]
            except IndexError:
                return None
        except ValueError:
            return getattr(self, str(item), None)

    def get_data(self, item=0):
        try:
            item_index = int(item)
            try:
                return self.data[item_index].data
            except IndexError:
                pass
                # fatal(f"No data found for index {item_index}. ")
        except ValueError:
            try:
                return getattr(self, item).data
            except AttributeError:
                pass
        fatal(f"No data found for item {item}. ")

    @property
    def data_is_none(self):
        return any([d is None or d.data_is_none for d in self.data])

    def _assert_data_is_not_none(self):
        if self.data_is_none:
            raise ValueError(
                "This data item does not contain any data yet. "
                "Use set_data() before applying any operations. "
            )

    def propagate_operator(self, other, operator):
        self._assert_data_is_not_none()
        if isinstance(other, (float, int)):
            new_data = [
                getattr(self.data[i], operator)(other)
                for i in range(self._tuple_length)
            ]
        else:
            new_data = [
                getattr(self.data[i], operator)(other.data[i])
                for i in range(self._tuple_length)
            ]
        return type(self)(self._data_item_classes, data=new_data)

    def propagate_operator_inplace(self, other, operator):
        self._assert_data_is_not_none()
        if isinstance(other, (float, int)):
            for i in range(self._tuple_length):
                getattr(self.data[i], operator)(other)
        else:
            for i in range(self._tuple_length):
                getattr(self.data[i], operator)(other.data[i])
        return self

    def __iadd__(self, other):
        return self.propagate_operator_inplace(other, "__iadd__")

    def __add__(self, other):
        return self.propagate_operator(other, "__add__")

    def __imul__(self, other):
        return self.propagate_operator_inplace(other, "__imul__")

    def __mul__(self, other):
        return self.propagate_operator(other, "__mul__")

    def __itruediv__(self, other):
        return self.propagate_operator_inplace(other, "__itruediv__")

    def __truediv__(self, other):
        return self.propagate_operator(other, "__truediv__")

    def inplace_merge_with(self, other):
        for i in range(self._tuple_length):
            # can only apply merge of both items exist (and contain data)
            if self.data[i] is not None and other.data[i] is not None:
                self.data[i].inplace_merge_with(other.data[i])
            else:
                # the case of both item None is acceptable
                # because the component not be activated in the actor, e.g. edep uncertainty,
                # but it should not occur that one item is None and the other is not.
                if (self.data[i] is None) is not (other.data[i] is None):
                    s_not = {True: "", False: "not"}
                    fatal(
                        "Cannot apply inplace merge data to container "
                        "with unset (None) data items. "
                        f"In this case, the inplace item {i} is {s_not[self.data[i] is None]} None, "
                        f"and the other item {i} is {s_not[other.data[i] is None]} None. "
                        f"This is likely an implementation error in GATE. "
                    )
        return self

    def merge_with(self, other):
        data = []
        for i in range(self._tuple_length):
            if (
                self.data[i] is not None
                and other.data[i] is not None
                and self.data[i].data is not None
                and other.data[i].data is not None
            ):
                data.append(self.data[i].merge_with(other.data[i]))
            else:
                # FIXME: we need a consistency check here
                data.append(None)

        return type(self)(
            self._data_item_classes,
            data=data,
        )

    def write(self, path, item, **kwargs):
        data_item = self.get_data_item_object(item)
        if data_item is not None:
            data_item.write(path)
        else:
            warning(f"Cannot write item {item} because it does not exist (=None).")

    def __getattr__(self, item):
        # check if any of the data items has this attribute
        # exclude 'data' to avoid infinite recursion
        # exclude '__setstate__' and '__getstate__' to avoid interference with pickling
        if item not in ("data", "__setstate__", "__getstate__"):
            methods_in_data = []
            attributes_in_data = []
            for d in self.data:
                if d is not None and hasattr(d, item):
                    if callable(getattr(d, item)):
                        methods_in_data.append(getattr(d, item))
                    else:
                        attributes_in_data.append(getattr(d, item))
            if len(attributes_in_data) > 0 and len(methods_in_data) > 0:
                fatal(
                    f"Cannot hand down request for attribute to data items "
                    f"because some contain it as a method and other as a property. "
                )
            elif len(attributes_in_data) > 0:
                if len(attributes_in_data) == 1:
                    return attributes_in_data[0]
                else:
                    return attributes_in_data
            elif len(methods_in_data) > 0:

                def hand_down(*args, **kwargs):
                    return_values = []
                    for m in methods_in_data:
                        return_values.append(m(*args, **kwargs))
                    return tuple(return_values)

                return hand_down
            else:
                raise AttributeError(f"No such attribute '{item}'")
        else:
            raise AttributeError(f"No such attribute '{item}'")


class SingleArray(DataItemContainer):

    _data_item_classes = (ArrayDataItem,)

    def __init__(self, *args, **kwargs):
        # specify the data item classes
        super().__init__(*args, **kwargs)


class DoubleArray(DataItemContainer):

    _data_item_classes = (ArrayDataItem, ArrayDataItem)

    def __init__(self, *args, **kwargs):
        # specify the data item classes
        super().__init__(*args, **kwargs)


class SingleItkImage(DataItemContainer):

    _data_item_classes = (ItkImageDataItem,)

    @property
    def image(self):
        return self.data[0].image


class SingleMeanItkImage(DataItemContainer):

    _data_item_classes = (MeanItkImageDataItem,)


class SingleItkImageWithVariance(DataItemContainer):

    _data_item_classes = (
        ItkImageDataItem,
        ItkImageDataItem,
    )

    # # Only the linear quantity is active by default
    # # the uncertainty quantity has write_to_disk=True by default so whenever it is activated,
    # # the results will be written to disk (probably the expected default behavior in most cases)
    # default_data_item_config = Box(
    #     {
    #         0: Box({"output_filename": "auto", "write_to_disk": True, "active": True}),
    #         1: Box(
    #             {"output_filename": "auto", "write_to_disk": False, "active": False}
    #         ),
    #         "variance": Box(
    #             {"output_filename": "auto", "write_to_disk": False, "active": False}
    #         ),
    #         "std": Box(
    #             {"output_filename": "auto", "write_to_disk": False, "active": False}
    #         ),
    #         "uncertainty": Box(
    #             {"output_filename": "auto", "write_to_disk": True, "active": False}
    #         ),
    #     }
    # )
    # let the class know which properties should be treated as data items
    __extra_data_items__ = ("variance", "std", "uncertainty")

    def get_variance_or_uncertainty(self, which_quantity):
        try:
            number_of_samples = self.data[0].number_of_samples
            value_array = np.asarray(self.data[0].data)
            if not number_of_samples > 1:
                warning(
                    "You try to compute statistical errors with only one or zero event! "
                    "The uncertainty value for all voxels has been fixed at 1"
                )
                output_arr = np.ones_like(value_array)
            elif self.data[1] is None or self.data[1].data is None:
                warning(
                    "This data item does not contain squared values so no variance can be calculated. "
                    "The variance will be set to 0 everywhere. "
                )
                output_arr = np.zeros_like(value_array)
            else:
                squared_value_array = np.asarray(self.data[1].data)
                output_arr = calculate_variance(
                    value_array, squared_value_array, number_of_samples
                )
                if which_quantity in (
                    "std",
                    "uncertainty",
                ):
                    output_arr = np.sqrt(output_arr)
                if which_quantity in ("uncertainty",):
                    output_arr = np.divide(
                        output_arr,
                        value_array / number_of_samples,
                        out=np.zeros_like(output_arr),
                        where=value_array != 0,
                    )
            output_image = itk_image_from_array(output_arr)
            output_image.CopyInformation(self.data[0].data)
        except AttributeError as e:
            fatal(str(e))
        return self._data_item_classes[0](data=output_image)

    @property
    def variance(self):
        return self.get_variance_or_uncertainty("variance")

    @property
    def std(self):
        return self.get_variance_or_uncertainty("std")

    @property
    def uncertainty(self):
        return self.get_variance_or_uncertainty("uncertainty")


class QuotientItkImage(DataItemContainer):

    _data_item_classes = (
        ItkImageDataItem,
        ItkImageDataItem,
    )

    # # Specify which items should be written to disk and how
    # # Important: define this at the class level, NOT in the __init__ method
    # default_data_item_config = Box(
    #     {
    #         0: Box({"output_filename": "auto", "write_to_disk": True, "active": True}),
    #         1: Box({"output_filename": "auto", "write_to_disk": True, "active": True}),
    #         "quotient": Box(
    #             {"output_filename": "auto", "write_to_disk": True, "active": True}
    #         ),
    #     }
    # )

    # let the class know which properties should be treated as data items
    __extra_data_items__ = ("quotient",)

    @property
    def quotient(self):
        return self.data[0] / self.data[1]


class QuotientMeanItkImage(QuotientItkImage):

    _data_item_classes = (
        MeanItkImageDataItem,
        MeanItkImageDataItem,
    )


def merge_data(list_of_data):
    merged_data = type(list_of_data[0])(
        list_of_data[0].belongs_to, data=list_of_data[0].data
    )
    for d in list_of_data[1:]:
        merged_data.inplace_merge_with(d)
    return merged_data


available_data_container_classes = {
    "SingleItkImage": SingleItkImage,
    "SingleMeanItkImage": SingleMeanItkImage,
    "QuotientMeanItkImage": QuotientMeanItkImage,
    "SingleArray": SingleArray,
    "DoubleArray": DoubleArray,
    "SingleItkImageWithVariance": SingleItkImageWithVariance,
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itk
import numpy as np
from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility
from opengate.contrib.beamlines.ionbeamline import BeamlineModel
from opengate.contrib.tps.ionbeamtherapy import spots_info_from_txt

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test044_pbs", output_folder="test059"
    )
    output_path = paths.output
    ref_path = paths.output_ref

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 12365478910
    sim.random_engine = "MersenneTwister"
    sim.output_dir = output_path

    # units
    km = gate.g4_units.km
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    deg = gate.g4_units.deg
    rad = gate.g4_units.rad

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    sim.world.size = [600 * cm, 500 * cm, 500 * cm]

    # nozzle box
    box = sim.add_volume("Box", "box")
    box.size = [500 * mm, 500 * mm, 1000 * mm]
    box.translation = [1148 * mm, 0.0, 0.0]
    box.rotation = Rotation.from_euler("y", -90, degrees=True).as_matrix()
    box.material = "Vacuum"
    box.color = [0, 0, 1, 1]

    # nozzle WET
    nozzle = sim.add_volume("Box", "nozzle")
    nozzle.mother = box.name
    nozzle.size = [500 * mm, 500 * mm, 2 * mm]
    nozzle.material = "G4_WATER"

    # target
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [500 * mm, 500 * mm, 400 * mm]
    phantom.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    phantom.translation = [-200.0, 0.0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    # roos chamber
    roos = sim.add_volume("Tubs", "roos")
    roos.mother = phantom.name
    roos.material = "G4_WATER"
    roos.rmax = 7.8
    roos.rmin = 0
    roos.dz = 200
    roos.color = [1, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = (
        "FTFP_INCLXX_EMZ"  # 'QGSP_BIC_HP_EMZ' #"FTFP_INCLXX_EMZ"
    )

    sim.physics_manager.set_production_cut("world", "all", 1000 * km)

    # add dose actor
    dose = sim.add_actor("DoseActor", "doseInXYZ")
    dose.output_filename = "abs_dose_roos.mhd"
    dose.attached_to = roos.name
    dose.size = [1, 1, 800]
    dose.spacing = [15.6, 15.6, 0.5]
    dose.hit_type = "random"
    dose.dose.active = True

    # ---------- DEFINE BEAMLINE MODEL -------------#
    IR2HBL = BeamlineModel()
    IR2HBL.name = None
    IR2HBL.radiation_types = "ion 6 12"
    # Nozzle entrance to Isocenter distance
    IR2HBL.distance_nozzle_iso = 1300.00  # 1648 * mm#1300 * mm
    # SMX to Isocenter distance
    IR2HBL.distance_stearmag_to_isocenter_x = 6700.00
    # SMY to Isocenter distance
    IR2HBL.distance_stearmag_to_isocenter_y = 7420.00
    # polinomial coefficients
    IR2HBL.energy_mean_coeffs = [11.91893485094217, -9.539517997860457]
    IR2HBL.energy_spread_coeffs = [0.0004790681841295621, 5.253257865904452]
    IR2HBL.sigma_x_coeffs = [2.3335753978880014]
    IR2HBL.theta_x_coeffs = [0.0002944903217664001]
    IR2HBL.epsilon_x_coeffs = [0.0007872786903040108]
    IR2HBL.sigma_y_coeffs = [1.9643343053823967]
    IR2HBL.theta_y_coeffs = [0.0007911780133478402]
    IR2HBL.epsilon_y_coeffs = [0.0024916149017600447]

    # --------START PENCIL BEAM SCANNING---------- #
    # NOTE: HBL means that the beam is coming from -x (90 degree rot around y)
    nSim = 50000  # 328935  # particles to simulate per beam
    beam_data_dict = spots_info_from_txt(
        ref_path / "TreatmentPlan4Gate-F5x5cm_E120MeVn.txt", "ion 6 12", beam_nr=1
    )
    tps = sim.add_source("TreatmentPlanPBSource", "TPSource")
    tps.n = nSim
    tps.sorted_spot_generation = True
    tps.beam_model = IR2HBL
    tps.beam_data_dict = beam_data_dict
    tps.beam_nr = 1
    tps.particle = "ion 6 12"
    ntot = beam_data_dict["msw_beam"]

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")

    # create output dir, if it doesn't exist
    output_path.mkdir(parents=True, exist_ok=True)

    # start simulation
    sim.run()

    # -------------END SCANNING------------- #
    # print results at the end
    print(stats)

    # ------ TESTS -------#
    dose_path = utility.scale_dose(
        str(dose.get_output_path("dose")),
        ntot / nSim,
        output_path / "abs_dose_roos-Scaled.mhd",
    )

    # ABSOLUTE DOSE

    # read output and ref
    f = ref_path / "idc-PHANTOM-roos-F5x5cm_E120MeVn-PLAN-Physical.mhd"
    print("Compare", dose_path, f)
    img_mhd_out = itk.imread(dose_path)
    img_mhd_ref = itk.imread(f)
    data = itk.GetArrayViewFromImage(img_mhd_out)
    data_ref = itk.GetArrayViewFromImage(img_mhd_ref)
    shape = data.shape
    spacing = img_mhd_out.GetSpacing()
    spacing_ref = np.flip(img_mhd_ref.GetSpacing())

    ok = utility.assert_img_sum(img_mhd_out, img_mhd_ref, sum_tolerance=5.5)

    print("compare dose at points")
    points = 400 - np.linspace(10, 14, 9)
    ok = (
        utility.compare_dose_at_points(
            points,
            data,
            data_ref,
            shape,
            data_ref.shape,
            spacing,
            spacing_ref,
            axis1="z",
            axis2="x",
            rel_tol=0.065,
        )
        and ok
    )

    # 1D
    """fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(25, 10))
    plot_img_axis(ax, img_mhd_out, "x profile", axis="z")
    plot_img_axis(ax, img_mhd_ref, "x ref", axis="z")
    plt.show()
    fig.savefig(output_path / "dose_profiles_water.png")"""

    utility.test_ok(ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import gatetools.phsp


def root_load_xyz(root_file: str, keys: [str]):
    data_ref, keys_ref, m_ref = gatetools.phsp.load(root_file)

    index_x = keys_ref.index(keys[0])
    index_y = keys_ref.index(keys[1])
    index_z = keys_ref.index(keys[2])

    xs = [data_ref_i[index_x] for data_ref_i in data_ref]
    ys = [data_ref_i[index_y] for data_ref_i in data_ref]
    zs = [data_ref_i[index_z] for data_ref_i in data_ref]

    return xs, ys, zs


if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test010_generic_source_thetaphi", "test010"
    )

    print(paths.output_ref)

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    deg = gate.g4_units.deg
    um = gate.g4_units.um
    g_cm3 = gate.g4_units.g_cm3

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.number_of_threads = 1
    sim.random_seed = 123654
    sim.output_dir = paths.output

    # materials
    sim.volume_manager.material_database.add_material_weights(
        "Vacuum", ["H"], [1], 1e-9 * g_cm3
    )

    # set the world size like in the Gate macro
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "Vacuum"

    # add a simple volume
    phsp = sim.add_volume("Box", "phsp")
    phsp.size = [40 * cm, 40 * cm, 1 * um]
    phsp.material = "Vacuum"

    # test sources
    source = sim.add_source("GenericSource", "beam")
    source.particle = "gamma"
    source.activity = 1e6 * Bq / sim.number_of_threads
    source.position.type = "point"
    source.position.translation = [0 * cm, 0 * cm, 1 * m]
    source.direction.type = "iso"
    source.direction.theta = [0 * deg, 10 * deg]
    source.direction.phi = [0 * deg, 360 * deg]
    source.energy.type = "mono"
    source.energy.mono = 1 * MeV

    # actors
    stats_actor = sim.add_actor("SimulationStatisticsActor", "Stats")

    phsp_actor = sim.add_actor("PhaseSpaceActor", "phspActor")
    phsp_actor.output_filename = "test010-thetaphi-phsp.root"
    phsp_actor.attached_to = "phsp"
    phsp_actor.attributes = [
        "Position",
    ]

    # verbose
    sim.g4_commands_after_init.append("/tracking/verbose 0")

    # start simulation
    sim.run()

    # print
    print("Simulation seed:", sim.current_random_seed)

    # get results
    print(stats_actor)

    # gate_test10_thetaphi
    # Gate mac/main.mac
    # Current version is two times faster (:
    print("-" * 80)

    g9_xs, g9_ys, g9_zs = root_load_xyz(
        str(paths.output_ref / "test010_thetaphi_phsp.root"), ["X", "Y", "Z"]
    )
    g9_xmin, g9_xmax = min(g9_xs), max(g9_xs)
    g9_ymin, g9_ymax = min(g9_ys), max(g9_ys)
    g9_zmin, g9_zmax = min(g9_zs), max(g9_zs)

    g10_xs, g10_ys, g10_zs = root_load_xyz(
        phsp_actor.get_output_path(), ["Position_X", "Position_Y", "Position_Z"]
    )
    g10_xmin, g10_xmax = min(g10_xs), max(g10_xs)
    g10_ymin, g10_ymax = min(g10_ys), max(g10_ys)
    g10_zmin, g10_zmax = min(g10_zs), max(g10_zs)

    # Tolerance in mm
    is_ok = True
    is_ok = is_ok and utility.check_diff_abs(
        g9_xmin, g10_xmin, tolerance=0.05, txt="x min"
    )
    is_ok = is_ok and utility.check_diff_abs(
        g9_xmax, g10_xmax, tolerance=0.05, txt="x max"
    )
    is_ok = is_ok and utility.check_diff_abs(
        g9_ymin, g10_ymin, tolerance=0.05, txt="y min"
    )
    is_ok = is_ok and utility.check_diff_abs(
        g9_ymax, g10_ymax, tolerance=0.055, txt="y max"
    )
    is_ok = is_ok and utility.check_diff_abs(
        g9_zmin, g10_zmin, tolerance=1e-2, txt="z min"
    )
    is_ok = is_ok and utility.check_diff_abs(
        g9_zmax, g10_zmax, tolerance=1e-2, txt="z max"
    )

    utility.test_ok(is_ok)



import os
import shutil
import zipfile
import requests
import colored
from pathlib import Path
from .g4DataSetup import *

color_error = colored.fore("red") + colored.style("bold")


# Check and download opengate tests data if not present:
def check_tests_data_folder():
    data_location = get_tests_data_folder()
    if not data_location.exists():
        print("No Opengate test data available in: " + str(data_location))
        print("I download it for you.")
        download_tests_data(data_location)
        print("")
        print("Done")
    else:
        # Check if the commit is correct if file HEAD is present
        if (data_location.parent / "HEAD").exists():
            with (data_location.parent / "HEAD").open("r") as f:
                checkout_reference_data_git = f.readline().strip()
            if (data_location / "sha.log").exists():
                with (data_location / "sha.log").open("r") as f:
                    checkout_real_data_git = f.readline().strip()
                if checkout_reference_data_git != checkout_real_data_git:
                    shutil.rmtree(data_location)
                    print(
                        "No correct Opengate test data version in: "
                        + str(data_location)
                    )
                    print("I update it for you.")
                    download_tests_data(data_location)
                    print("")
                    print("Done")
            else:
                shutil.rmtree(data_location)
                print("No Opengate test data available in: " + str(data_location))
                print("I download it for you.")
                download_tests_data(data_location)
                print("")
                print("Done")
        # Check if the size of one .raw file is correct to detect lfs
        if (data_location / "ct_4mm.raw").is_file():
            filesize = (data_location / "ct_4mm.raw").stat().st_size
            if filesize < 4000000:
                print(
                    "It seems the test data in: "
                    + str(data_location)
                    + " do not have the correct size"
                )
                print("Maybe you do not have git-lfs. Execute this:")
                print("Install git-lfs from https://git-lfs.com/")
                print("cd " + str(data_location))
                print("git-lfs pull")
                return False
        else:  # if the file is not present
            print(
                colored.stylize(
                    "The data are not present in: " + str(data_location),
                    color_error,
                )
            )
            print("Download them with:")
            print("git submodule update --init --recursive")
            return False
    return True


# Download opengate tests data:
def download_tests_data(data_location: Path):
    data_location.mkdir(parents=True, exist_ok=True)
    with (data_location.parent / "HEAD").open("r") as f:
        checkout_data_git = str(f.readline()).strip()
    url = (
        "https://gitlab.in2p3.fr/api/v4/projects/15155/repository/commits/"
        + checkout_data_git
    )
    r = requests.get(url=url)
    js = r.json()
    id_pipeline = str(js["last_pipeline"]["id"])
    url = (
        "https://gitlab.in2p3.fr/api/v4/projects/15155/pipelines/"
        + id_pipeline
        + "/jobs"
    )
    r = requests.get(url=url)
    js = r.json()[0]
    id_pipeline = str(js["id"])
    url = (
        "https://gitlab.in2p3.fr/api/v4/projects/15155/jobs/"
        + id_pipeline
        + "/artifacts"
    )

    # download the archive (with resume if the connexion failed)
    filename = url.split("/")[-1]
    out = data_location / filename
    download_with_resume(url, out)

    if filename == "artifacts":
        with zipfile.ZipFile(data_location / filename, "r") as zip_ref:
            zip_ref.extractall(data_location)
    os.remove(data_location / filename)

    with zipfile.ZipFile(
        data_location / "artifact_zip" / f"{checkout_data_git}.zip",
        "r",
    ) as zip_ref:
        zip_ref.extractall(str(data_location))
    shutil.rmtree(data_location / "artifact_zip", ignore_errors=True)


# Return opengate tests data folder:
def get_tests_data_folder() -> Path:
    package_location = Path(__file__).resolve().parent
    data_location = package_location.parent / "opengate" / "tests" / "data"
    for parent in package_location.parents:
        if (parent / "opengate").exists():
            data_location = parent / "opengate" / "tests" / "data"
            break
    if not data_location:
        print("Cannot find opengate folder near: " + str(package_location))
    return data_location



import sys
import time
import scipy
from scipy.spatial.transform import Rotation
import numpy as np
import threading
from box import Box
import itk
import bisect
import random

import opengate_core as g4
from .phspsources import PhaseSpaceSource
from ..exception import fatal
from .generic import GenericSource
from ..image import get_info_from_image
from ..image import compute_image_3D_CDF
from ..utility import LazyModuleLoader
from ..base import process_cls


#
torch = LazyModuleLoader("torch")
gaga = LazyModuleLoader("gaga_phsp")


def generate_isotropic_directions(
    n, min_theta=0, max_theta=np.pi, min_phi=0, max_phi=2 * np.pi, rs=np.random
):
    """
    like in G4SPSAngDistribution.cc

    Later : do a version with torch (gpu) instead of np (cpu) ?
    """
    u = rs.uniform(0, 1, size=n)
    costheta = np.cos(min_theta) - u * (np.cos(min_theta) - np.cos(max_theta))
    sintheta = np.sqrt(1 - costheta**2)

    v = rs.uniform(0, 1, size=n)
    phi = min_phi + (max_phi - min_phi) * v
    sinphi = np.sin(phi)
    cosphi = np.cos(phi)

    # "direct cosine" method, like in Geant4 (already normalized)
    px = -sintheta * cosphi
    py = -sintheta * sinphi
    pz = -costheta

    # concat
    v = np.column_stack((px, py, pz))

    return v


class VoxelizedSourcePDFSampler:
    """
    This is an alternative to GateSPSVoxelsPosDistribution (c++)
    It is needed because the cond voxel source is used on python side.

    There are two versions, version 2 is much slower (do not use)
    """

    def __init__(self, itk_image, version=1):
        self.image = itk_image
        self.version = version
        # get image in np array
        self.imga = itk.array_view_from_image(itk_image)
        imga = self.imga

        # image sizes
        lx = self.imga.shape[0]
        ly = self.imga.shape[1]
        lz = self.imga.shape[2]
        m = lx * ly * lz

        # normalized pdf
        pdf = imga.ravel(order="F")
        self.pdf = pdf / pdf.sum()

        # create grid of indices
        [x_grid, y_grid, z_grid] = np.meshgrid(
            np.arange(lx), np.arange(ly), np.arange(lz), indexing="ij"
        )

        # list of indices
        self.xi, self.yi, self.zi = (
            x_grid.ravel(order="F"),
            y_grid.ravel(order="F"),
            z_grid.ravel(order="F"),
        )

        # 1D indices
        self.linear_indices = np.arange(int(m))

        if version == 2:
            self.init_cdf()

        if version == 3:
            self.init_cdf()

        # ------------------------------------------
        # Below = NOT WORKING, DO NOT USE
        # create grid of indices in physical space
        """img_info = gate.get_info_from_image(self.image)
        sp = img_info.spacing
        hsp = sp / 2
        s = img_info.size * sp
        [px_grid, py_grid, pz_grid] = np.meshgrid(
            np.linspace(hsp[0], s[0] - hsp[0], lx),
            np.linspace(hsp[1], s[1] - hsp[1], lx),
            np.linspace(hsp[2], s[2] - hsp[2], lz),
            indexing='ij',
        )
        # list of indices
        self.pxi, self.pyi, self.pzi = (
            px_grid.ravel(order='F'),
            py_grid.ravel(order='F'),
            pz_grid.ravel(order='F'),
        )
        """
        # ------------------------------------------

    def init_cdf(self):
        self.cdf_x, self.cdf_y, self.cdf_z = compute_image_3D_CDF(self.image)
        self.cdf_x = np.array(self.cdf_x)
        self.cdf_y = np.array(self.cdf_y)
        self.cdf_z = np.array(self.cdf_z)
        self.cdf_init = True

    def searchsorted2d(self, a, b):
        # https://stackoverflow.com/questions/56471109/how-to-vectorize-with-numpy-searchsorted-in-a-2d-array
        # Inputs : a is (m,n) 2D array and b is (m,) 1D array.
        # Finds np.searchsorted(a[i], b[i])) in a vectorized way by
        # scaling/offsetting both inputs and then using searchsorted

        # Get scaling offset and then scale inputs
        s = np.r_[0, (np.maximum(a.max(1) - a.min(1) + 1, b) + 1).cumsum()[:-1]]
        a_scaled = (a + s[:, None]).ravel()
        b_scaled = b + s

        # Use searchsorted on scaled ones and then subtract offsets
        return np.searchsorted(a_scaled, b_scaled) - np.arange(len(s)) * a.shape[1]

    def sample_indices_slower(self, n, rs=np.random):
        """
        This version seems slower than the other version with np random choice
        """
        # Z (here, search sorted is faster than bisect+loop)
        uz = rs.uniform(0, 1, size=n)
        # i = [bisect.bisect_left(self.cdf_z, uz[t], lo=0, hi=lz) for t in range(n)]
        i = np.searchsorted(self.cdf_z, uz, side="left")

        # Y, knowing Z
        # https://stackoverflow.com/questions/56471109/how-to-vectorize-with-numpy-searchsorted-in-a-2d-array
        ly = self.imga.shape[1]
        uy = rs.uniform(0, 1, size=n)
        j = [bisect.bisect_left(self.cdf_y[i[t]], uy[t], lo=0, hi=ly) for t in range(n)]

        # (here search sorted is not faster, we keep bisect)
        # slower:
        # cdfyi = np.take(self.cdf_y, i, axis=0)
        # j = self.searchsorted2d(cdfyi, uy)

        # X
        lx = self.imga.shape[2]
        ux = rs.uniform(0, 1, size=n)
        k = [
            bisect.bisect_left(self.cdf_x[i[t]][j[t]], ux[t], lo=0, hi=lx)
            for t in range(n)
        ]

        return i, j, k

    def samples_g4(self, n):
        # to compare with cpp version
        sps = g4.GateSPSVoxelsPosDistribution()
        sps.SetCumulativeDistributionFunction(self.cdf_z, self.cdf_y, self.cdf_x)
        p = np.array([sps.VGenerateOneDebug() for a in range(n)])
        return p[:, 2], p[:, 1], p[:, 0]

    def sample_indices(self, n, rs=np.random):
        indices = rs.choice(self.linear_indices, size=n, replace=True, p=self.pdf)
        i = self.xi[indices]
        j = self.yi[indices]
        k = self.zi[indices]
        return i, j, k

    def sample_indices_phys(self, n, rs=np.random):
        # TODO (not used yet)
        indices = rs.choice(self.linear_indices, size=n, replace=True, p=self.pdf)
        i = self.pxi[indices]
        j = self.pyi[indices]
        k = self.pzi[indices]
        return i, j, k


class VoxelizedSourceConditionGenerator:
    def __init__(
        self, activity_source_filename, rs=np.random, use_activity_origin=False
    ):
        self.activity_source_filename = str(activity_source_filename)
        # options
        self.compute_directions = False
        self.use_activity_origin = use_activity_origin
        self.translation = [0, 0, 0]
        self.rotation = Rotation.identity().as_matrix()
        # variables
        self.image = None
        self.cdf_x = self.cdf_y = self.cdf_z = None
        self.rs = rs
        self.source_img_info = None
        self.sampler = None
        self.points_offset = None
        # init
        self.is_initialized = False

    def initialize_source(self):
        # FIXME warning, this is call in the same thread but several time (???)
        if self.image is None:
            # sleep to avoid concurrent access (??)
            time.sleep(random.uniform(0.3, 1.0))
            self.image = itk.imread(self.activity_source_filename)
        self.source_img_info = get_info_from_image(self.image)
        if self.sampler is None:
            self.sampler = VoxelizedSourcePDFSampler(self.image)
        self.rs = np.random
        # we set the points in the g4 coord system (according to the center of the image)
        # or according to the activity source image origin
        if self.use_activity_origin is True:
            self.points_offset = self.source_img_info.origin
        else:
            hs = self.source_img_info.spacing / 2.0
            self.points_offset = -hs * self.source_img_info.size + hs
        self.is_initialized = True

    def generate_condition(self, n):
        if self.is_initialized is False:
            self.initialize_source()

        # i j k is in np array order = z y x
        # but img_info is in the order x y z
        i, j, k = self.sampler.sample_indices(n, self.rs)

        # half pixel size
        hs = self.source_img_info.spacing / 2.0

        # sample within the voxel
        rx = self.rs.uniform(-hs[0], hs[0], size=n)
        ry = self.rs.uniform(-hs[1], hs[1], size=n)
        rz = self.rs.uniform(-hs[2], hs[2], size=n)

        # warning order np is z,y,x while itk is x,y,z
        x = self.source_img_info.spacing[2] * k + rz
        y = self.source_img_info.spacing[1] * j + ry
        z = self.source_img_info.spacing[0] * i + rx

        # x,y,z are in the image coord system
        # they are offset according to the coord system (image center or image offset)
        p = np.column_stack((x, y, z)) + self.points_offset + self.translation

        # rotation
        p = np.dot(p, self.rotation.T)

        # need direction ?
        if self.compute_directions is False:
            return p
        v = generate_isotropic_directions(n, rs=self.rs)
        v = np.dot(v, self.rotation.T)
        return np.column_stack((p, v))


class GANSource(GenericSource, g4.GateGANSource):
    """
    GAN source: the Generator produces particles
    Input is a neural network Generator trained with a GAN
    """

    user_info_defaults = {
        "pth_filename": (
            None,
            {"doc": "Filename of the Generator (.pth), train with gaga_train"},
        ),
        "backward_distance": (
            None,
            {
                "doc": "Change position of the generated particle according to this distance in backward direction (-direction) "
            },
        ),
        "backward_force": (
            None,
            {
                "doc": "if backward is enabled and the time is not managed by the GAN the time cannot be changed (yet). Use 'force' to enable backward if you know what your are doing"
            },
        ),
        "energy_min_threshold": (
            -1,
            {
                "doc": "Minimum energy threshold. If the generator creates particles with less than this energy, "
                "the particle is skipped"
            },
        ),
        "energy_max_threshold": (
            sys.float_info.max,
            {
                "doc": "Maximum energy threshold. If the generator creates particles with a higher energy, "
                "the particle is skipped"
            },
        ),
        "batch_size": (
            10000,
            {
                "doc": "Batch size of the GAN generated particles. ",
            },
        ),
        "position_keys": (
            None,
            {"doc": "FIXME"},
        ),
        "direction_keys": (
            None,
            {"doc": "FIXME"},
        ),
        "energy_key": (
            None,
            {"doc": "FIXME"},
        ),
        "time_key": (
            None,
            {"doc": "FIXME"},
        ),
        "weight_key": (
            None,
            {"doc": "FIXME"},
        ),
        "relative_timing": (
            True,
            {"doc": "FIXME"},
        ),
        "generator": (
            None,
            {"doc": "FIXME"},
        ),
        "verbose_generator": (
            False,
            {
                "doc": "Print information every time the generator generates a batch of particles. "
            },
        ),
        "use_time": (
            False,
            {"doc": "Consider the particle time given by the generator. "},
        ),
        "use_weight": (
            False,
            {"doc": "Consider the particle weight given by the generator. "},
        ),
        "cond_image": (
            None,
            {
                "doc": "Filename of the activity distribution (provided as image) to use for the conditional GAN"
            },
        ),
        "cond_debug": (
            False,
            {"doc": "Print debug information for conditional GAN"},
        ),
        "compute_directions": (
            False,
            {"doc": "FIXME"},
        ),
        "skip_policy": (
            "SkipEvents",
            {
                "doc": "When some generated events are skipped (e.g. with the energy threshold), they could be "
                "either 'skipped' (SkippedEvents) or generated with energy=zero (ZeroEnergy) to preserve "
                "the correct number of events",
                "allowed_values": ("SkipEvents", "ZeroEnergy"),
            },
        ),
        "gpu_mode": (
            "gpu",
            {
                "doc": "Use the GPU or CPU for the GAN, or let GATE decide. ",
                "allowed_values": ("auto", "cpu", "gpu"),
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateGANSource.__init__(self)

    def initialize(self, run_timing_intervals):
        # FIXME -> check input user_info
        # initialize the mother class generic source
        GenericSource.initialize(self, run_timing_intervals)

        # default generator or set by the user
        if self.user_info.generator is None:
            self.set_default_generator()
        gen = self.user_info.generator

        # initialize the generator (read the GAN)
        # this function must have 1) the generator function 2) the associated info
        gen.initialize()

        # set the function pointer to the cpp side
        self.SetGeneratorFunction(gen.generator)

        # set the parameters to the cpp side
        self.SetGeneratorInfo(gen.gan_info)

    def set_default_generator(self):
        # non-conditional generator
        if self.cond_image is None:
            self.generator = GANSourceDefaultGenerator(self.user_info)
            return

        # FIXME: I changed this line because the second arg 'self' seemed wrong to me. Check!
        vcg = VoxelizedSourceConditionGenerator(self.cond_image)
        # vcg = VoxelizedSourceConditionGenerator(self.cond_image, self)
        vcg.compute_directions = self.compute_directions
        self.generator = GANSourceConditionalGenerator(
            self.user_info, vcg.generate_condition
        )


class GANPairsSource(GANSource, g4.GateGANPairSource):
    """
    GAN source: the Generator produces pairs of particles (for PET)
    Input is a neural network Generator trained with a GAN
    """

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateGANPairSource.__init__(self)

    def set_default_generator(self):
        # non-conditional generator
        if self.user_info.cond_image is None:
            self.user_info.generator = GANSourceDefaultPairsGenerator(self.user_info)
            return

        # conditional generator
        fatal(
            f"A conditional generator must be set in the "
            f"user_info.generator option of the GANPairsSource '{self.user_info.name}'."
        )


class GANSourceDefaultGenerator:
    """
    This class manage the base components of a particle generator.

    - In the constructor, the module 'gaga' is imported. It is only imported in the constructor to only required
    this module if it is used

    - 'initialize' function: the GAN is loaded and the list of keys is initialized

    - 'generator' function: default generator

    - 'get_output_keys' function: map the user defined keys to the ones of the generator. There are two usages, either
    with on single primary (3 values for position, direction), or paired primary (6 values).

    - 'move_backward' function: consider all particles positions and move their backward according to their direction,
    using the factor provided by the user in 'user_info.backward_distance'. This is useful to allow generating
    particles that do not intersect with the detector.

    - 'copy_generated_particle_to_g4' function: copy all the particles (pos, dir, time, energy) to the cpp part.

    """

    def __init__(self, user_info):
        self.user_info = user_info
        # self.gaga = None
        self.indexes_are_build = None
        self.lock = None
        self.initialize_is_done = False
        self.keys_output = None
        self.gan_info = None
        self.gpu_mode = None

    def __getstate__(self):
        self.lock = None
        # self.gaga = None
        self.gan_info = None
        return self.__dict__

    def initialize(self):
        self.lock = threading.Lock()
        with self.lock:
            self.gpu_mode = self.user_info.gpu_mode
            if not self.initialize_is_done:
                self.read_gan_and_keys()
                self.initialize_is_done = True

    def read_gan_and_keys(self):
        # allow converting str like 1e5 to int
        self.user_info.batch_size = int(float(self.user_info.batch_size))

        # FIXME check the number of params

        # read pth and create the gan info structure
        self.gan_info = Box()
        g = self.gan_info
        g.params, g.G, g.D, g.optim = gaga.load(
            self.user_info.pth_filename, self.gpu_mode
        )

        """
        gan_info structure
        - params    net params from gaga_phsp (including the keys)
        - G         Generator net
        - D         Discriminator net
        - optim     Info about net optimisation

        We analyse the keys and fill the following elements to initialize the GANSource
        - info.position_is_set_by_GAN
        - info.direction_is_set_by_GAN
        - info.energy_is_set_by_GAN
        - info.time_is_set_by_GAN
        - info.weight_is_set_by_GAN
        - info.timing_is_relative
        """

        # by default, the output keys are the same as the input keys
        # (could be changed later by some parameterization)
        if self.keys_output is None:
            g.params.keys_output = g.params.keys_list
        else:
            g.params.keys_output = self.keys_output

        # position,
        g.position_is_set_by_GAN = self.user_info.position_keys is not None
        g.direction_is_set_by_GAN = self.user_info.direction_keys is not None
        g.energy_is_set_by_GAN = self.user_info.energy_key is not None
        g.time_is_set_by_GAN = self.user_info.time_key is not None
        g.weight_is_set_by_GAN = self.user_info.weight_key is not None
        g.timing_is_relative = self.user_info.relative_timing

        # init the key index
        g.position_gan_index = -1
        g.direction_gan_index = -1
        g.energy_gan_index = -1
        g.time_gan_index = -1
        g.weight_gan_index = -1

    def get_output_keys(self):
        with self.lock:
            # (this is only performed once)
            if not self.indexes_are_build:
                self.get_output_keys_with_lock()
                self.indexes_are_build = True

    def get_output_keys_with_lock(self):
        n = self.user_info.batch_size
        g = self.gan_info
        the_keys = g.params.keys_output

        self.check_parameters(g)
        self.get_position_index(g, the_keys, n)
        self.get_direction_index(g, the_keys, n)
        self.get_energy_index(g, the_keys, n)
        self.get_time_index(g, the_keys, n)
        self.get_weight_index(g, the_keys, n)

    def check_parameters(self, g):
        # position
        if g.position_is_set_by_GAN and len(self.user_info.position_keys) != 3:
            dim = len(self.user_info.position_keys)
            self.fatal(f"you must provide 3 values for position, while it was {dim}")
        # direction
        if g.direction_is_set_by_GAN and len(self.user_info.direction_keys) != 3:
            dim = len(self.user_info.direction_keys)
            self.fatal(f"you must provide 3 values for direction, while it was {dim}")

    def fatal(self, txt):
        fatal(f"Error in the GANSource {self.user_info.name}: {txt}")

    def get_position_index(self, g, the_keys, n):
        # get position index from GAN (or a fixed value)
        if not g.position_is_set_by_GAN:
            return
        g.position_gan_index, g.position_use_index = self.get_gan_key_index(
            the_keys, self.user_info.position_keys, n
        )

    def get_direction_index(self, g, the_keys, n):
        # get position from GAN (or a fixed value)
        if not g.direction_is_set_by_GAN:
            return
        g.direction_gan_index, g.direction_use_index = self.get_gan_key_index(
            the_keys, self.user_info.direction_keys, n
        )

    def get_energy_index(self, g, the_keys, n):
        # get energy index from GAN
        if not g.energy_is_set_by_GAN:
            return
        g.energy_gan_index = the_keys.index(self.user_info.energy_key)

    def get_time_index(self, g, the_keys, n):
        # get time index from GAN
        if not g.time_is_set_by_GAN:
            return
        g.time_gan_index = the_keys.index(self.user_info.time_key)

    def get_weight_index(self, g, the_keys, n):
        # get weight index from GAN
        if not g.weight_is_set_by_GAN:
            return
        g.weight_gan_index = the_keys.index(self.user_info.weight_key)

    def get_gan_key_index(self, all_keys, user_keys, n):
        """
        Consider 'user_keys' in the list of all keys and return the indexes
        Special case: if the key name is a float value, we consider the value instead of the index.
        """
        p = []
        o = []  # true or false if used or not
        dim = len(user_keys)
        for i in range(dim):
            try:
                index = all_keys.index(user_keys[i])
                p.append(index)
                o.append(True)
            except:
                if type(user_keys[i]) == float:
                    p.append(np.ones(n) * user_keys[i])
                    o.append(False)
                else:
                    fatal(
                        f"Error, cannot use the key {user_keys[i]} (in {user_keys}) in the GAN source. "
                        f"GAN keys are: {all_keys}"
                    )
        return p, o

    def generator(self, source):
        """
        Main function that will be called from the cpp side every time a batch
        of particles should be created.
        Once created here, the particles are copied to cpp.
        (Yes maybe the copy could be avoided, but I did not manage to do it)
        """
        # get the info
        g = self.gan_info
        n = self.user_info.batch_size
        start = None

        # verbose and timing ?
        if self.user_info.verbose_generator:
            start = time.time()
            print(f"Generate {n} particles from GAN ", end="")

        # generate samples (this is the most time-consuming part)
        fake = gaga.generate_samples_non_cond(
            g.params,
            g.G,
            n=n,
            batch_size=n,
            normalize=False,
            to_numpy=True,
            silence=True,
        )

        # consider the names of the output keys position/direction/energy/time/weight
        self.get_output_keys()

        # move particle backward ?
        self.move_backward(g, fake)

        # copy to cpp
        self.copy_generated_particle_to_g4(source, g, fake)

        # verbose
        if self.user_info.verbose_generator:
            end = time.time()
            print(f"in {end - start:0.1f} sec (GPU={g.params.current_gpu_mode})")

    def copy_generated_particle_to_g4(self, source, g, fake):
        # get the index of from the GAN vector
        # (or some fixed values)

        # position
        if g.position_is_set_by_GAN:
            pos = []
            dim = len(g.position_gan_index)
            for i in range(dim):
                if g.position_use_index[i]:
                    pos.append(fake[:, g.position_gan_index[i]])
                else:
                    pos.append(g.position_gan_index[i])
            # copy to c++
            source.fPositionX = pos[0]
            source.fPositionY = pos[1]
            source.fPositionZ = pos[2]

        # direction
        if g.direction_is_set_by_GAN:
            dir = []
            dim = len(g.direction_gan_index)
            for i in range(dim):
                if g.direction_use_index[i]:
                    dir.append(fake[:, g.direction_gan_index[i]])
                else:
                    dir.append(g.direction_gan_index[i])
            # copy to c++
            source.fDirectionX = dir[0]
            source.fDirectionY = dir[1]
            source.fDirectionZ = dir[2]

        # energy
        if g.energy_is_set_by_GAN:
            # copy to c++
            source.fEnergy = fake[:, g.energy_gan_index]

        # time
        if g.time_is_set_by_GAN:
            # copy to c++
            source.fTime = fake[:, g.time_gan_index]

        # weight
        if g.weight_is_set_by_GAN:
            # copy to c++
            source.fWeight = fake[:, g.weight_gan_index]

    def move_backward(self, g, fake):
        # move particle backward ?
        back = self.user_info.backward_distance
        if not back:
            return

        if not g.time_is_set_by_GAN and not self.user_info.backward_force:
            fatal(
                f"If backward is enabled the time is not managed by GAN,"
                f" time is wrong. IT can be forced, however, with the option 'backward_force'"
            )

        # move particle position
        position = fake[:, g.position_gan_index[0] : g.position_gan_index[0] + 3]
        direction = fake[:, g.direction_gan_index[0] : g.direction_gan_index[0] + 3]
        fake[:, g.position_gan_index[0] : g.position_gan_index[0] + 3] = (
            position - back * direction
        )

        # modify the time because we move the particle backward
        if g.time_is_set_by_GAN:
            c = scipy.constants.speed_of_light * 1000 / 1e9  # in mm/ns
            xt = back / c
            fake[:, g.time_gan_index] -= xt


class GANSourceDefaultPairsGenerator(GANSourceDefaultGenerator):
    """
    Like GANSourceDefaultGenerator but for pairs of particle (PET)
    """

    def __init__(self, user_info):
        super().__init__(user_info)
        self.is_paired = True

    def __getstate__(self):
        self.lock = None
        self.gan_info = None
        return self.__dict__

    def check_parameters(self, g):
        # position
        if g.position_is_set_by_GAN and len(self.user_info.position_keys) != 6:
            dim = len(self.user_info.position_keys)
            self.fatal(f"you must provide 6 values for position, while it was {dim}")
        # direction
        if g.direction_is_set_by_GAN and len(self.user_info.direction_keys) != 6:
            dim = len(self.user_info.direction_keys)
            self.fatal(f"you must provide 6 values for direction, while it was {dim}")

    def get_energy_index(self, g, the_keys, n):
        # get energy index from GAN
        if not g.energy_is_set_by_GAN:
            return
        ek = self.user_info.energy_key
        dim = len(ek)
        if dim != 2:
            self.fatal(f"you must provide 2 values for energy, while it was {dim}")
        g.energy_gan_index = [the_keys.index(ek[0]), the_keys.index(ek[1])]

    def get_time_index(self, g, the_keys, n):
        # get time index from GAN
        if not g.time_is_set_by_GAN:
            return
        ek = self.user_info.time_key
        dim = len(ek)
        if dim != 2:
            self.fatal(f"you must provide 2 values for time, while it was {dim}")
        g.time_gan_index = [the_keys.index(ek[0]), the_keys.index(ek[1])]

    def get_weight_index(self, g, the_keys, n):
        # get weight index from GAN
        if not g.weight_is_set_by_GAN:
            return
        ek = self.user_info.weight_key
        dim = len(ek)
        if dim != 2:
            self.fatal(f"you must provide 2 values for weight, while it was {dim}")
        g.weight_gan_index = [the_keys.index(ek[0]), the_keys.index(ek[1])]

    def generator(self, source):
        """
        Main function that will be called from the cpp side every time a batch
        of particles should be created.
        Once created here, the particles are copied to cpp.
        (Yes maybe the copy could be avoided, but I did not manage to do it)
        """
        # get the info
        g = self.gan_info
        n = self.user_info.batch_size
        start = None

        # verbose and timing ?
        if self.user_info.verbose_generator:
            start = time.time()
            print(f"Generate {n} particles from GAN ", end="")

        # generate samples (this is the most time-consuming part)
        fake = gaga.generate_samples_non_cond(
            g.params,
            g.G,
            n=n,
            batch_size=n,
            normalize=False,
            to_numpy=True,
            silence=True,
        )

        # consider the names of the output keys position/direction/energy/time/weight
        self.get_output_keys()

        # move particle backward ?
        self.move_backward(g, fake)

        # copy to cpp
        self.copy_generated_particle_to_g4(source, g, fake)

        # verbose
        if self.user_info.verbose_generator:
            end = time.time()
            print(f"in {end - start:0.1f} sec (device={g.params.current_gpu_device})")

    def copy_generated_particle_to_g4(self, source, g, fake):
        # position
        if g.position_is_set_by_GAN:
            pos = []
            dim = len(g.position_gan_index)
            for i in range(dim):
                if g.position_use_index[i]:
                    pos.append(fake[:, g.position_gan_index[i]])
                else:
                    pos.append(g.position_gan_index[i])
            # copy to c++
            source.fPositionX = pos[0]
            source.fPositionY = pos[1]
            source.fPositionZ = pos[2]
            source.fPositionX2 = pos[3]
            source.fPositionY2 = pos[4]
            source.fPositionZ2 = pos[5]

        # direction
        if g.direction_is_set_by_GAN:
            dir = []
            dim = len(g.direction_gan_index)
            for i in range(dim):
                if g.direction_use_index[i]:
                    dir.append(fake[:, g.direction_gan_index[i]])
                else:
                    dir.append(g.direction_gan_index[i])
            # copy to c++
            source.fDirectionX = dir[0]
            source.fDirectionY = dir[1]
            source.fDirectionZ = dir[2]
            source.fDirectionX2 = dir[3]
            source.fDirectionY2 = dir[4]
            source.fDirectionZ2 = dir[5]

        # energy
        if g.energy_is_set_by_GAN:
            # copy to c++
            source.fEnergy = fake[:, g.energy_gan_index[0]]
            source.fEnergy2 = fake[:, g.energy_gan_index[1]]

        # time
        if g.time_is_set_by_GAN:
            # copy to c++
            source.fTime = fake[:, g.time_gan_index[0]]
            source.fTime2 = fake[:, g.time_gan_index[1]]

        # weight
        if g.weight_is_set_by_GAN:
            # copy to c++
            source.fWeight = fake[:, g.weight_gan_index[0]]
            source.fWeight2 = fake[:, g.weight_gan_index[1]]

    def move_backward(self, g, fake):
        # move particle backward ?
        back = self.user_info.backward_distance
        if not back:
            return

        if not g.time_is_set_by_GAN and not self.user_info.backward_force:
            fatal(
                f"If backward is enabled the time is not managed by GAN,"
                f" time is wrong. IT can be forced, however, with the option 'backward_force'"
            )

        # move particle position
        position = fake[:, g.position_gan_index[0] : g.position_gan_index[0] + 3]
        direction = fake[:, g.direction_gan_index[0] : g.direction_gan_index[0] + 3]
        fake[:, g.position_gan_index[0] : g.position_gan_index[0] + 3] = (
            position - back * direction
        )

        # move second particle position
        position = fake[:, g.position_gan_index[3] : g.position_gan_index[3] + 3]
        direction = fake[:, g.direction_gan_index[3] : g.direction_gan_index[3] + 3]
        fake[:, g.position_gan_index[3] : g.position_gan_index[3] + 3] = (
            position - back * direction
        )

        # modify the time because we move the particle backward
        if g.time_is_set_by_GAN:
            c = scipy.constants.speed_of_light * 1000 / 1e9  # in mm/ns
            xt = back / c
            fake[:, g.time_gan_index[0]] -= xt
            fake[:, g.time_gan_index[1]] -= xt


class GANSourceConditionalGenerator(GANSourceDefaultGenerator):
    def __init__(self, user_info, generate_condition_function):
        super().__init__(user_info)
        self.generate_condition = generate_condition_function

    def generate_condition(self, n):
        fatal(
            f'Error: to use GANSourceConditionalGenerator,  you must provide a function "f" '
            f'that take a single int "n" as input and generate n condition samples. '
            f'This function "f" must be set with generator.generate_condition = f'
        )
        return None

    def generator(self, source):
        """
        Generate particles with a GAN, considering conditional vectors.
        """

        # get the info
        g = self.gan_info
        n = self.user_info.batch_size
        start = None

        # verbose and timing ?
        if self.user_info.verbose_generator:
            start = time.time()
            print(f"Generate {n} particles from GAN ", end="")

        # generate cond
        cond = self.generate_condition(n)

        # generate samples (this is the most time-consuming part)
        if self.user_info.cond_debug:
            # debug : do not run GAN, only consider the conditions
            # needed by test 047
            fake = cond
        else:
            fake = gaga.generate_samples3(
                g.params,
                g.G,
                n=n,
                cond=cond,
            )

        # consider the names of the output keys position/direction/energy/time/weight
        self.get_output_keys()

        # if debug, the GAN is not used.
        if self.user_info.cond_debug:
            g.position_gan_index = [0, 1, 2]
            g.direction_gan_index = [3, 4, 5]
            g.energy_is_set_by_GAN = False
            g.energy_type = False

        # move particle backward ?
        self.move_backward(g, fake)

        # copy to cpp (g4)
        self.copy_generated_particle_to_g4(source, g, fake)

        # verbose
        if self.user_info.verbose_generator:
            end = time.time()
            print(f"in {end - start:0.2f} sec (GPU={g.params.current_gpu_mode})")


class GANSourceConditionalPairsGenerator(GANSourceDefaultPairsGenerator):
    """
    Generate pairs of particles with a GAN, considering conditional vectors.

    The parameter sphere_radius is required : it is the radius of the sphere
    that surround the source during the training.

    """

    def __init__(self, user_info, sphere_radius, generate_condition_function):
        super().__init__(user_info)
        self.sphere_radius = sphere_radius
        self.generate_condition = generate_condition_function

    def __getstate__(self):
        # needed to not pickle. Need to reset some attributes
        self.gan = None
        self.generate_condition = None
        self.lock = None
        return self.__dict__

    def generate_condition(self, n):
        fatal(
            f'Error: to use GANSourceConditionalPairsGenerator,  you must provide a function "f" '
            f'that take a single int "n" as input and generate n condition samples. '
            f'This function "f" must be set with generator.generate_condition = f'
        )
        return None

    def generator(self, source):
        # get the info
        g = self.gan_info
        n = self.user_info.batch_size
        start_time = None

        # verbose and timing ?
        if self.user_info.verbose_generator:
            start_time = time.time()
            print(f"Generate {n} particles from GAN ", end="")

        # generate cond
        cond = self.generate_condition(n)

        # generate samples (this is the most time-consuming part)
        fake = gaga.generate_samples3(g.params, g.G, to_numpy=False, n=n, cond=cond)

        # parametrisation
        keys = g.params["keys_list"]
        if self.sphere_radius is None:
            fatal(
                f'Error the option "sphere_radius" must be set for the source "{self.user_info.name}"'
            )
        params = {
            "keys_list": keys,
            "radius": self.sphere_radius,
            "ignore_directions": False,
        }
        fake = gaga.from_tlor_to_pairs(fake, params)
        keys = params["keys_output"]
        g.params["keys_output"] = keys

        # consider the names of the output keys position/direction/energy/time/weight
        self.get_output_keys()

        # move particle backward ?
        self.move_backward(g, fake)

        # back from torch to numpy
        fake = fake.cpu().data.numpy()

        # copy to cpp
        self.copy_generated_particle_to_g4(source, g, fake)

        # verbose
        if self.user_info.verbose_generator:
            end = time.time()
            print(
                f"in {end - start_time:0.1f} sec (device={g.params.current_gpu_device})"
            )


process_cls(GANSource)
process_cls(GANPairsSource)



from loguru import logger
import os
import sys

# Ensure UTF-8 encoding is used
os.environ["PYTHONUTF8"] = "1"
if sys.getdefaultencoding() != "utf-8":
    sys.stdin.reconfigure(encoding="utf-8")
    sys.stdout.reconfigure(encoding="utf-8")
    sys.stderr.reconfigure(encoding="utf-8")


# Remove the default logger configuration
logger.remove()

# configure logger
logger.level("DEBUG", color="<green>")
logger.level("DEBUG", icon="🐞 ")
logger.level("INFO", icon="")
logger.level("INFO", color="<green>")
logger.level("NONE", no=100, color="")
logger.level("WARNING", icon="⚠️ ")
logger.level("CRITICAL", icon="☠️ ")

# default log level of loguru
DEBUG = 10
INFO = 20
WARNING = 30
CRITICAL = 50
NONE = 100

# default log level for running_verbose_level
RUN = 20
EVENT = 50


def log_level(log_handler_id):
    handler_config = logger._core.handlers.get(log_handler_id)
    return handler_config.levelno



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import uproot


def test_half_life_fit(sim, output, half_life, ax):
    sec = gate.g4_units.s
    keV = gate.g4_units.keV

    print(f"root file {output}")

    root = uproot.open(output)
    branch_name = root.keys()[0]
    branch = root[branch_name]["GlobalTime"]
    time = branch.array(library="numpy") / sec
    branch = root[branch_name]["KineticEnergy"]
    E = branch.array(library="numpy")

    # consider time of arrival for both peaks
    time1 = time[(E > 150 * keV) & (E < 250 * keV)]
    print(f"Total number of gammas           = {len(time1)}")

    tmax = sim.run_timing_intervals[1][1] / sec
    time2 = time1[time1 < tmax]
    print(f"Number of gammas from 0 to {tmax} sec = {len(time2)}")

    # fit for half life
    start_time = sim.run_timing_intervals[0][0] / sec
    end_time = sim.run_timing_intervals[0][1] / sec
    hl, xx, yy = utility.fit_exponential_decay(time2, start_time, end_time)
    # compare with source half_life (convert in sec)
    tol = 0.09
    hl_ref = half_life / sec
    diff = abs(hl - hl_ref) / hl_ref
    b = diff < tol
    diff *= 100
    utility.print_test(b, f"Half life {hl_ref:.2f} sec vs {hl:.2f} sec : {diff:.2f}% ")

    ax.hist(
        time2,
        bins=100,
        label="decay source",
        histtype="stepfilled",
        alpha=0.5,
        range=(0, sim.run_timing_intervals[1][1] / sec + 10),
        density=True,
    )
    ax.plot(xx, yy, label="fit half-life {:.2f} sec".format(hl))
    ax.legend()
    ax.set_xlabel("time (s)")
    ax.set_ylabel("produced gammas")

    return b, len(time2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test010_confine")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.number_of_threads = 1
    sim.output_dir = paths.output

    # some units
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    m = gate.g4_units.m
    deg = gate.g4_units.deg
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    Bq = gate.g4_units.Bq

    # set the world size like in the Gate macro
    sim.world.size = [1 * m, 1 * m, 1 * m]

    # add a simple volume
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 0 * cm]
    waterbox.material = "G4_WATER"

    # volume where to confine
    stuff = sim.add_volume("Cons", "stuff")
    stuff.mother = "waterbox"
    stuff.rmin1 = 0
    stuff.rmax1 = 0.5 * cm
    stuff.rmin2 = 0
    stuff.rmax2 = 0.5 * cm
    stuff.dz = 2 * cm
    stuff.dphi = 360 * deg
    stuff.translation = [-5 * cm, 0 * cm, 0 * cm]
    stuff.material = "G4_WATER"

    # daughter volume
    stuffi = sim.add_volume("Cons", "stuff_inside")
    stuffi.mother = stuff.name
    stuffi.rmin1 = 0
    stuffi.rmax1 = 0.4 * cm
    stuffi.rmin2 = 0
    stuffi.rmax2 = 0.4 * cm
    stuffi.dz = 2 * cm
    stuffi.dphi = 360 * deg
    stuffi.translation = [-0.1 * cm, 0 * cm, 0 * cm]
    stuffi.material = "G4_AIR"

    # activity
    activity = 500000 * Bq
    # activity = 50 * Bq

    # test confined source
    source = sim.add_source("GenericSource", "non_confined_src")
    source.attached_to = "stuff"
    source.particle = "gamma"
    source.activity = activity / sim.number_of_threads
    source.position.type = "box"
    source.position.size = [5 * cm, 5 * cm, 5 * cm]
    source.direction.type = "momentum"
    source.direction.momentum = [-1, 0, 0]
    source.energy.type = "mono"
    source.energy.mono = 1 * MeV

    # test confined source
    """
       the source is confined in the given volume ('stuff'), it means that
       all particles will be emitted only in this volume.
       The 'box' type is required to define a larger volume than 'stuff'.
       It is done here by computing the bounding box
       Daughter volumes of 'stuff' do not count : no particle will be generated
       from 'stuff_inside'
    """
    source = sim.add_source("GenericSource", "confined_src")
    source.attached_to = "stuff"
    source.particle = "gamma"
    source.activity = activity / sim.number_of_threads
    source.position.type = "box"
    source.position.size = sim.volume_manager.volumes[
        source.attached_to
    ].bounding_box_size
    print("Source size", source.position.size)
    pMin, pMax = sim.volume_manager.volumes[source.attached_to].bounding_limits
    source.position.confine = "stuff"
    source.direction.type = "momentum"
    source.direction.momentum = [1, 0, 0]
    source.energy.type = "mono"
    source.energy.mono = 1 * MeV

    # actors
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")

    dose_actor = sim.add_actor("DoseActor", "dose_actor")
    dose_actor.output_filename = "test010-2.mhd"
    # dose_actor.output_filename = paths.output_ref / 'test010-2-edep.mhd'
    dose_actor.attached_to = waterbox
    dose_actor.size = [100, 100, 100]
    dose_actor.spacing = [2 * mm, 1 * mm, 1 * mm]

    # start simulation
    sim.run()

    # print
    print("Simulation seed:", sim.current_random_seed)

    # get results
    print(stats)

    # tests
    stats_ref = utility.read_stat_file(paths.output_ref / "test010_confine_stats.txt")
    is_ok = utility.assert_stats(stats, stats_ref, 0.10)
    is_ok = is_ok and utility.assert_images(
        paths.output_ref / "test010-2-edep.mhd",
        dose_actor.edep.get_output_path(),
        stats,
        tolerance=59,
        ignore_value_data2=0,
        apply_ignore_mask_to_sum_check=False,  # force legacy behavior
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import uproot
import matplotlib.pyplot as plt
import opengate as gate
import opengate.contrib.phantoms.nemaiec as gate_iec
from opengate.tests import utility


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", output_folder="test058")

    # create the simulation
    sim = gate.Simulation()
    simu_name = "main0_orientation"

    # options
    sim.number_of_threads = 1
    sim.visu = False
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = True
    sim.random_seed = 123456789
    sim.output_dir = paths.output
    sim.progress_bar = True

    # units
    m = gate.g4_units.m
    keV = gate.g4_units.keV
    cm = gate.g4_units.cm
    cm3 = gate.g4_units.cm3
    Bq = gate.g4_units.Bq
    BqmL = Bq / cm3

    # world size
    sim.world.size = [0.4 * m, 0.4 * m, 0.4 * m]

    # add IEC phantom
    iec1 = gate_iec.add_iec_phantom(sim, name="iec", check_overlap=True)

    # bg source
    s = gate_iec.add_background_source(sim, "iec", "bg", 100 * BqmL, True)
    s.particle = "gamma"
    s.energy.type = "mono"
    s.energy.mono = 10 * keV

    # spheres source
    sources = gate_iec.add_spheres_sources(
        sim, "iec", "spheres", "all", [100 * BqmL] * 6, True
    )
    for s in sources:
        s.particle = "e-"
        s.energy.type = "mono"
        s.energy.mono = 1 * keV

    # phys
    sim.physics_manager.set_production_cut("world", "all", 100 * m)

    # stats
    stats = sim.add_actor("SimulationStatisticsActor", "stats")

    # phsp
    phsp_bg = sim.add_actor("PhaseSpaceActor", "phsp_bg")
    phsp_bg.attributes = ["EventPosition"]
    phsp_bg.output_filename = "iec_bg.root"
    phsp_bg.steps_to_store = "first"
    f = sim.add_filter("ParticleFilter", "g")
    f.particle = "gamma"
    f.policy = "accept"
    phsp_bg.filters.append(f)

    phsp_sph = sim.add_actor("PhaseSpaceActor", "phsp_sph")
    phsp_sph.attributes = ["EventPosition"]
    phsp_sph.output_filename = "iec_spheres.root"
    phsp_sph.steps_to_store = "first"
    f = sim.add_filter("ParticleFilter", "electron")
    f.particle = "e-"
    f.policy = "accept"
    phsp_sph.filters.append(f)
    print(phsp_sph)

    print(sim.filter_manager.available_filters)
    print(sim.filter_manager.dump())

    # run
    sim.run()

    # end
    print(stats)

    # read root bg
    root = uproot.open(phsp_bg.get_output_path())
    tree = root[root.keys()[0]]
    posx = tree["EventPosition_X"].array()
    posy = tree["EventPosition_Y"].array()
    posz = tree["EventPosition_Z"].array()

    # consider only points around the sphere's centers
    # index = (posz > 3.5 * cm) & (posz < 3.9 * cm)
    index = (posz > 2 * cm) & (posz < 3.5 * cm)
    posx = posx[index]
    posy = posy[index]

    f, ax = plt.subplots(1, 1, figsize=(15, 5))
    ax.scatter(posx, posy, s=1)
    plt.gca().set_aspect("equal")

    plt.tight_layout()
    file = paths.output / "test058_bg.pdf"
    plt.savefig(file, bbox_inches="tight", format="pdf")
    print(f"Output plot is {file}")

    # read root sph
    root = uproot.open(phsp_sph.get_output_path())
    tree = root[root.keys()[0]]
    posx = tree["EventPosition_X"].array()
    posy = tree["EventPosition_Y"].array()
    posz = tree["EventPosition_Z"].array()

    # consider only points around the sphere's centers
    # index = (posz > 3.5 * cm) & (posz < 3.9 * cm)
    index = (posz > 2 * cm) & (posz < 3.5 * cm)
    posx = posx[index]
    posy = posy[index]

    f, ax = plt.subplots(1, 1, figsize=(15, 5))
    ax.scatter(posx, posy, s=1)
    plt.gca().set_aspect("equal")

    plt.tight_layout()
    file = paths.output / "test058_spheres.pdf"
    plt.savefig(file, bbox_inches="tight", format="pdf")
    print(f"Output plot is {file}")

    # ref root
    ref_root_file = paths.output_ref / "iec_bg.root"
    k = ["EventPosition_X", "EventPosition_Y", "EventPosition_Z"]
    is_ok = utility.compare_root3(
        ref_root_file,
        phsp_bg.get_output_path(),
        "phsp_bg",
        "phsp_bg",
        k,
        k,
        [0.27] * len(k),
        [1] * len(k),
        [1] * len(k),
        paths.output / "test058_bg.png",
    )

    # ref root
    print()
    ref_root_file = paths.output_ref / "iec_spheres.root"
    k = ["EventPosition_X", "EventPosition_Y", "EventPosition_Z"]
    is_ok = (
        utility.compare_root3(
            ref_root_file,
            phsp_sph.get_output_path(),
            "phsp_sph",
            "phsp_sph",
            k,
            k,
            [1.5] * len(k),
            [1] * len(k),
            [1] * len(k),
            paths.output / "test058_spheres.png",
        )
        and is_ok
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import test028_ge_nm670_spect_4_acc_angle_helpers as test028
from opengate.tests import utility

if __name__ == "__main__":
    # create the simulation
    sim = gate.Simulation()

    # simu description
    spect, proj = test028.create_spect_simu(
        sim,
        test028.paths,
        number_of_threads=3,
        activity_kBq=1000,
        aa_enabled=True,
        aa_mode="SkipEvents",
        version="_4_acc_angle_aa_se_mt",
    )

    # go
    sim.run()

    # check
    is_ok = test028.compare_result(
        sim,
        proj,
        "test028_aa_skip_events.png",
        sum_tolerance=25,
        version="_4_acc_angle_aa_se_mt",
    )
    utility.test_ok(is_ok)



import os
import re
import sys
import platform
import subprocess
import json
import setuptools
import sysconfig
from pathlib import Path


def warning(s):
    print(s)


def get_cmake_dir() -> Path:
    plat_name = sysconfig.get_platform()
    python_version = sysconfig.get_python_version()
    dir_name = f"cmake.{plat_name}-{sys.implementation.name}-{python_version}"
    cmake_dir = Path(get_base_dir()) / "core" / "build" / dir_name
    cmake_dir.mkdir(parents=True, exist_ok=True)
    return cmake_dir


def get_base_dir() -> Path:
    return Path(__file__).parent.parent.resolve()


with open("../VERSION", "r") as fh:
    version = fh.read()[:-1]

from setuptools import Extension, find_packages
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r"version\s*([\d.]+)", out.decode()).group(1)
            )
            if cmake_version < "3.1.0":
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        # required for auto-detection of auxiliary "native" libs

        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
            "-DPYTHON_EXECUTABLE=" + sys.executable,
        ]

        # cfg = "Debug" if self.debug else 'Release'
        cfg = "Release"
        build_args = ["--config", cfg]

        # Pile all .so in one place and use $ORIGIN as RPATH
        cmake_args += ["-DCMAKE_BUILD_WITH_INSTALL_RPATH=TRUE"]
        cmake_args += ["-DCMAKE_INSTALL_RPATH_USE_LINK_PATH=TRUE"]
        cmake_args += ["-DCMAKE_INSTALL_RPATH={}".format("$ORIGIN")]
        # cmake_args += ['-DCMAKE_CXX_FLAGS="-Wno-self-assign -Wno-extra-semi"']
        cmake_args += ["-DCMAKE_EXPORT_COMPILE_COMMANDS=ON"]

        if platform.system() == "Windows":
            cmake_args += [
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
            ]
            # cmake_args += ['-G "CodeBlocks - NMake Makefiles"']
            if sys.maxsize > 2**32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:
            cmake_args += ['-DCMAKE_CXX_FLAGS="-Wno-pedantic"']
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += ["--", "-j4"]

        env = os.environ.copy()

        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        print()
        print()
        warning("-----------------------------------")
        warning("-----------------------------------")
        print()
        print()

        print(f"Build folder : {self.build_temp}")

        print("Try to open config.json file")
        fn = "config.json"
        sconfig = {"G4INSTALL": "", "ITKDIR": ""}
        if "G4INSTALL" in env:
            sconfig["G4INSTALL"] = env["G4INSTALL"]
        if "ITKDIR" in env:
            sconfig["ITKDIR"] = env["ITKDIR"]
        try:
            f = open(fn, "r")
            sconfig = json.load(f)
        except IOError:
            print("No config file, use default")
        print("Config : ", sconfig)
        print()

        cmake_args += ["-DGeant4_DIR=" + sconfig["G4INSTALL"]]
        cmake_args += ["-DITK_DIR=" + sconfig["ITKDIR"]]

        print("CMAKE args", cmake_args)
        print()

        cmake_dir = get_cmake_dir()
        print("CMAKE build dir", cmake_dir)
        print()

        print("Starting cmake ...")
        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=cmake_dir, env=env
        )
        print("cmake done")
        # subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=cmake_dir)

        subprocess.check_call(
            ["cmake", "--build", ".", "--target", ext.name] + build_args,
            cwd=cmake_dir,
        )


if platform.system() == "Darwin":
    package_data = {
        "opengate_core": ["plugins/platforms/*.dylib"]
        + ["plugins/imageformats/*.dylib"]
    }
    # package_data = {}
else:
    package_data = {"opengate_core": ["plugins/*/*.so"] + ["plugins/*.so"]}

setuptools.setup(
    name="opengate-core",
    version=version,
    author="Opengate collaboration",
    author_email="david.sarrut@creatis.insa-lyon.fr",
    description="Simulation for Medical Physics",
    long_description="Simulation for Medical Physics",
    long_description_content_type="text/markdown",
    url="https://github.com/OpenGATE/opengate",
    ext_package="opengate_core",
    ext_modules=[CMakeExtension("opengate_core")],
    cmdclass=dict(build_ext=CMakeBuild),
    packages=find_packages(),
    package_data=package_data,
    zip_safe=False,
    python_requires=">=3.9",
    include_package_data=True,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
    install_requires=["wget", "colored>1.5", "requests"],
)



import tables
import numpy as np

from opengate.postprocessors.image import (
    create_projection_from_positions,
    write_image_with_pytables,
)

# ******
# This test is work in progress to try the experimental postprocessor subpackage.
# Disregard for the moment.
# ******


n_rows = 1000
x = np.random.randn(n_rows) * 20
y = np.random.randn(n_rows) * 20
z = np.ones_like(x)
positions = np.vstack((x, y, z)).T
size = [100, 100, 1]
spacing = [1, 1, 1]

output_file_handle = tables.open_file("test.h5", "w")
hdf5_group = output_file_handle.create_group("/", "projector")

projection = create_projection_from_positions(positions, size, spacing)
path_to_image = ".test072_projection.h5"
with tables.open_file(path_to_image, "w") as f:
    group_path_in_image_file = write_image_with_pytables(projection, file_handle=f)
    output_file_handle.create_external_link(
        hdf5_group, "proj_test", f"{str(path_to_image)}:{group_path_in_image_file}"
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from test077_scatter_helpers import *

if __name__ == "__main__":
    # paths
    paths = utility.get_default_test_paths(__file__, output_folder="test077_scatter")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.number_of_threads = 1
    sim.random_seed = 1321654
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    deg = gate.g4_units.deg
    MeV = gate.g4_units.MeV

    # world
    world = sim.world
    world.size = [5 * m, 5 * m, 5 * m]
    world.material = "G4_AIR"

    # waterbox
    wb = sim.add_volume("Box", "waterbox")
    wb.size = [20 * cm, 30 * cm, 30 * cm]
    wb.material = "G4_WATER"

    # detector
    det = sim.add_volume("Box", "detector")
    det.size = [10 * cm, 60 * cm, 60 * cm]
    det.translation = [40 * cm, 0, 0]
    det.material = "G4_WATER"
    det.material = "G4_Ir"

    # phys
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.set_production_cut("waterbox", "all", 0.1 * mm)

    # source
    source = sim.add_source("GenericSource", "beam")
    source.particle = "gamma"
    source.energy.type = "gauss"
    source.energy.mono = 0.5 * MeV
    source.energy.sigma_gauss = 0  # .2 * MeV
    source.position.type = "box"
    source.position.translation = [-60 * cm, 0, 0]
    source.position.size = [0 * cm, 8 * cm, 8 * cm]
    # source.direction.type = "focused"
    # source.direction.focus_point = [-20 * cm, 0, 0]
    source.direction.type = "momentum"
    source.direction.momentum = [1, 0, 0]
    source.n = 50000

    # stats
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True

    # phsp
    att_list = [
        "ParticleName",
        "ParentID",
        "PrePosition",
        "PostPosition",
        "EventKineticEnergy",
        "EventDirection",
        "EventID",
        "TrackID",
        "KineticEnergy",
        "PreDirection",
        "PostDirection",
        "UnscatteredPrimaryFlag",
    ]
    phsp = sim.add_actor("PhaseSpaceActor", "phsp")
    phsp.attached_to = det.name
    phsp.attributes = att_list
    # phsp.debug = True
    phsp.output_filename = "test077_scatter.root"
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "gamma"
    f.policy = "accept"
    phsp.filters.append(f)

    # phsp
    phsp2 = sim.add_actor("PhaseSpaceActor", "phsp_scatter")
    phsp2.attached_to = det.name
    phsp2.attributes = att_list
    phsp2.output_filename = phsp.output_filename
    # phsp2.debug = True
    fs = sim.add_filter("UnscatteredPrimaryFilter", "f_scatter")
    fs.policy = "accept"
    phsp2.filters.append(f)
    phsp2.filters.append(fs)

    # phsp
    phsp3 = sim.add_actor("PhaseSpaceActor", "phsp_no_scatter")
    phsp3.attached_to = det.name
    phsp3.attributes = att_list
    phsp3.output_filename = phsp.output_filename
    fs = sim.add_filter("UnscatteredPrimaryFilter", "f_no_scatter")
    fs.policy = "reject"
    phsp3.filters.append(f)
    phsp3.filters.append(fs)

    # start simulation
    sim.run()
    print(stats)

    # test
    print(phsp.get_output_path())
    print(phsp2.get_output_path())
    print(phsp3.get_output_path())
    is_ok = check_scatter(phsp.get_output_path())
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import click
import os
import opengate_core as g4

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("--input", "-i", default="", help="Input visualization file")
@click.option("--vrml", "-v", is_flag=True, help="VRML file")
@click.option("--gdml", "-g", is_flag=True, help="GDML file")
@click.option("--qt", "-q", is_flag=True, help="Test Qt window")
def go(input, vrml, gdml, qt):

    if qt:
        g4.createTestQtWindow()
        return ()

    if not os.path.isfile(input):
        print("The file " + input + " does not exist")
        return ()

    if vrml and gdml:
        print("Choose between vrml or gdml")
        return ()

    if vrml:
        try:
            import pyvista
        except:
            print(
                "The module pyvista is not installed to be able to visualize vrml files. Execute:"
            )
            print("pip install pyvista")
            return
        pl = pyvista.Plotter()
        pl.import_vrml(input)  # self.simulation.visu_filename)
        pl.add_axes(line_width=5)
        pl.show()

    if gdml:
        try:
            import pyg4ometry
        except:
            print(
                "The module pyg4ometry is not installed to be able to visualize gdml files. Execute:"
            )
            print("pip install pyg4ometry")
            return
        r = pyg4ometry.gdml.Reader(input)  # self.simulation.visu_filename)
        l = r.getRegistry().getWorldVolume()
        v = pyg4ometry.visualisation.VtkViewerColouredMaterial()
        v.addLogicalVolume(l)
        v.view()


# --------------------------------------------------------------------------
if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from scipy.spatial.transform import Rotation
import numpy as np
import uproot


# The test generates three different generic sources, momentum, focused and iso, which are attached to a plan.
# The plan is randomly rotated, and we verify that the generated particles have a direction which is in accordance
# with the applied rotations and the transmissions.


def test068(tab, nb_run, nb_part, nb_source):
    nb_event = len(tab)
    nb_event_theo = nb_run * nb_part * nb_source
    err = np.sqrt(nb_event_theo)
    if nb_event_theo - 3 * err < nb_event < nb_event_theo + 3 * err:
        return True
    else:
        return False


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test068")
    output_path = paths.output

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    # sim.running_verbose_level = gate.logger.EVENT
    sim.number_of_threads = 1
    sim.random_seed = 123654987
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    km = gate.g4_units.km
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    nm = gate.g4_units.nm
    Bq = gate.g4_units.Bq
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    sec = gate.g4_units.s
    gcm3 = gate.g4_units["g/cm3"]
    deg = gate.g4_units.deg

    #  adapt world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_Galactic"

    # nb of particles
    n = 100

    # Plan to attach the source
    nb_source = 3
    plan = sim.add_volume("Box", "source_plan")
    plan.mother = world.name
    plan.material = "G4_Galactic"
    plan.size = [10 * cm, 10 * cm, 1 * nm]
    plan.translation = np.array([0 * cm, 0 * cm, 0 * cm])

    source1 = sim.add_source("GenericSource", "photon_source")
    source1.particle = "gamma"
    source1.position.type = "box"
    source1.attached_to = plan.name
    source1.position.size = [10 * cm, 10 * cm, 1 * nm]
    source1.direction.type = "momentum"
    source1.direction_relative_to_attached_volume = True
    # source1.direction.focus_point = [0*cm, 0*cm, -5 *cm]
    source1.direction.momentum = [0, 0, -1]
    source1.energy.type = "mono"
    source1.energy.mono = 1 * MeV
    source1.activity = n * Bq / sim.number_of_threads

    source2 = sim.add_source("GenericSource", "photon_source_2")
    source2.particle = "gamma"
    source2.position.type = "box"
    source2.attached_to = plan.name
    source2.position.size = [10 * cm, 10 * cm, 1 * nm]
    source2.direction.type = "focused"
    source2.direction_relative_to_attached_volume = True
    source2.direction.focus_point = [0 * cm, 0 * cm, -5 * cm]
    # source1.direction.momentum = [0, 0, -1]
    source2.energy.type = "mono"
    source2.energy.mono = 1 * MeV
    source2.activity = n * Bq / sim.number_of_threads

    source3 = sim.add_source("GenericSource", "photon_source_3")
    source3.particle = "gamma"
    source3.position.type = "disc"
    source3.position.radius = 0 * mm
    source3.attached_to = plan.name
    source3.position.size = [10 * cm, 10 * cm, 1 * nm]
    source3.direction.type = "iso"
    source3.direction.theta = [0 * deg, 10 * deg]
    source3.direction.phi = [0 * deg, 360 * deg]
    source3.direction_relative_to_attached_volume = True
    source3.energy.type = "mono"
    source3.energy.mono = 1 * MeV
    source3.activity = n * Bq / sim.number_of_threads

    # Phase Space

    phsp_plan = sim.add_volume("Box", "phsp_plan")
    phsp_plan.mother = world.name
    phsp_plan.material = "G4_Galactic"
    phsp_plan.size = [10 * cm, 10 * cm, 1 * nm]
    phsp_plan.translation = np.array([0 * cm, 0 * cm, -5 * cm])

    phsp = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    phsp.attached_to = phsp_plan.name
    phsp.attributes = ["EventID"]
    phsp.output_filename = "test068.root"

    # MOTION for the source and actors
    sim.run_timing_intervals = []
    motion_source_rotations = []
    motion_source_translations = []
    motion_phsp_rotations = []
    motion_phsp_translations = []

    for i in range(10):
        x_translation = 20 * cm * np.random.random()
        y_translation = 20 * cm * np.random.random()
        z_translation = 20 * cm * np.random.random()

        x_rot = 360 * np.random.random()
        y_rot = 360 * np.random.random()
        z_rot = 360 * np.random.random()

        motion_source_translations.append([x_translation, y_translation, z_translation])

        rot = Rotation.from_euler(
            "xyz", [x_rot, y_rot, z_rot], degrees=True
        ).as_matrix()
        vec_source_phsp = phsp_plan.translation - plan.translation
        translation_phsp = np.dot(rot, vec_source_phsp) + np.array(
            [x_translation, y_translation, z_translation]
        )

        motion_source_rotations.append(rot)
        motion_phsp_rotations.append(rot)
        motion_phsp_translations.append(translation_phsp)

        sim.run_timing_intervals.append([i * sec, (i + 1) * sec])
    plan.add_dynamic_parametrisation(
        translation=motion_source_translations, rotation=motion_source_rotations
    )
    phsp_plan.add_dynamic_parametrisation(
        translation=motion_phsp_translations, rotation=motion_phsp_rotations
    )

    # stat actor
    s = sim.add_actor("SimulationStatisticsActor", "Stats")
    s.track_types_flag = True

    # Physic list and cuts
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.gamma = 1 * mm
    sim.physics_manager.global_production_cuts.electron = 1 * mm
    sim.physics_manager.global_production_cuts.positron = 1 * mm

    # go !
    sim.run()
    print(f"{phsp.get_output_path() = }")
    # Validation test
    f_phsp = uproot.open(phsp.get_output_path())
    arr = f_phsp["PhaseSpace"].arrays()
    print("Number of detected events :", len(arr))
    nb_event = len(arr)
    nb_part = n
    nb_run = len(sim.run_timing_intervals)
    err = np.sqrt(nb_part * nb_run * nb_source)
    print(
        "Number of expected events : ["
        + str(int(-3 * err + nb_run * nb_part * nb_source))
        + ","
        + str(int(+3 * err + nb_run * nb_part * nb_source))
        + "]"
    )
    is_ok = test068(arr, len(sim.run_timing_intervals), nb_part, nb_source)
    utility.test_ok(is_ok)
